//Potential.h

#pragma once

#ifndef _POTENTIAL_H_
#define _POTENTIAL_H_

#include "vec.h"
#include "glm.h"
#include "sparse_matrix.h"
using namespace std;

#define MAXgridX 128  
#define MAXgridY 128
#define	MAXgridZ 128
#define KNOWN 0    
#define CANDIDATE 1       
#define UNKNOWN 2
#define MININT 0.0001 
#define MAXINT 100000

struct list{
	int x, y ,z;
	list* next;
};

typedef struct{
	float speed;
	float cost;
	Vec3f velocity;
}faceDataStruct;

typedef struct{ 
	float discomfort; 
	float potential; //ܳ
	faceDataStruct faceData[6];  //ϵݣҪcost
	int FMMFlag;  //ǸĴ״̬UNKNOWN,CANDIDATE,KNOWN
	int HeapPosition; //ڶе
}gridDataStruct;

class Potential 
{
private:

	//cost·
	gridDataStruct m_gridData[MAXgridX][MAXgridY][MAXgridZ];
	//ܶȳȺܶȺϰγɵܶȳ
	float density[MAXgridX][MAXgridY][MAXgridZ]; 
	//ƽٶȳܶȳɣspeed
	Vec3f averagev[MAXgridX][MAXgridY][MAXgridZ];
	//ϰ볡ϰܶȳ
	float distance[MAXgridX][MAXgridY][MAXgridZ];
	
	float GridLength, DisValue;

	float maxSpeed, minSpeed;
	int m_GoalXa, m_GoalXb, m_GoalYa, m_GoalYb, m_GoalZa, m_GoalZb;
	int gridX, gridY, gridZ;

	float maxDensity, minDensity;
	float lamda;
	float w1, w2, w3;

	vector<int> Candidate; //FMM
	int heapSize;
	SparseMatrix<Vec3f> edge_normals;
	int tmpgridX, tmpgridY, tmpgridZ;
	int totalGrids;

public:

	Potential(Vec3f size, float _GridLength, float _DisValue);
	void SetGoal(Vec3f goalmin, Vec3f goalmax);
	void SetSpeedRange(float _minSpeed, float _maxSpeed);
	void SetDensityRange(float _minDensity, float _maxDensity);
	void SetLamdaWeights(float _lamda, float _w1, float _w2, float _w3);
	void resetAllGridData(void);
	void ComputeFaceData(void);
	void FMM(void);
	Vec3f GetVelocity(Vec3f position);
	void addDensity(Vec3f position,Vec3f velocity);
	bool AddDiscomfort(GLMmodel* model);
	bool isObstacle(Vec3f position);

private:

	void CalcPotential(int x, int y, int z);
	void solve(float &x, float a, float b, float c, float d);
	void process(int x, int y, int z);
	float ComputeDistance(GLMmodel* model, int trangle_index, int index_a,int index_b,int index_c,Vec3f trangle_point_a,Vec3f trangle_point_b,Vec3f trangle_point_c,Vec3f point);
	float inside_or_outside(Vec3f point, Vec3f signormal);
	bool Is_point_on_segment(Vec3f point,Vec3f pointA,Vec3f pointB);
	void Floodfill(int i, int j, int k);
	void Vertex_and_edge_Normals(GLMmodel* model,SparseMatrix<Vec3f>& edge_normals);
	Vec3f get_SparseMatrix_value(int pointA_index,int pointB_index);
	Vec3f get_PotentialGradient(Vec3f position);
	Vec3f get_DensityGradient(Vec3f position);

	void AddToHeap(int x, int y, int z);
	void UpdateHeap(int x, int y, int z);
	int PopHeap(void);
	inline int GI(int i, int j, int k) const { return (i * gridY + j) * gridZ + k; }
	inline void GIJK(int index, int& i, int& j, int& k)
	{ k = index % gridZ; j =index / gridZ;
	i = j / gridY;      j = j % gridY; }
	
};

#endif