#include "Simulator.h"

#ifdef _OPENMP
#include <omp.h>
#endif
#include "Agent.h"
#include "KdTree.h"

#include "CIniReader.h"
extern CIniReader *fin;

namespace RVO
{
Simulator::Simulator() : agents_(), defaultAgent_(NULL), globalTime_(0.0f), kdTree_(NULL), timeStep_(1.0f)
{
	kdTree_ = new KdTree(this);
}

Simulator::Simulator(float timeStep, float neighborDist, size_t maxNeighbors, float timeHorizon, float radius, float maxSpeed, const Vec3f& velocity) : agents_(), defaultAgent_(NULL), globalTime_(0.0f), kdTree_(NULL), timeStep_(timeStep)
{
	kdTree_ = new KdTree(this);
	defaultAgent_ = new Agent(this);

	defaultAgent_->maxNeighbors_ = maxNeighbors;
	defaultAgent_->maxSpeed_ = maxSpeed;
	defaultAgent_->neighborDist_ = neighborDist;
	defaultAgent_->radius_ = radius;
	defaultAgent_->timeHorizon_ = timeHorizon;
	defaultAgent_->velocity_ = velocity;
}

void Simulator::init()
{
	if (globalTime_ != 0)
	{
		reset_particles(fin->ReadInteger("noise","particlenum"));
		return;
	}
	makenoise = new MakeNoise();

	seed = fin->ReadInteger("noise","seed");
	initial_min = fin->ReadVec3f("noise","initial_min");
	initial_max = fin->ReadVec3f("noise","initial_max");
	sphere_centre1 = fin->ReadVec3f("noise","sphere_centre1");
	sphere_radius1 = fin->ReadFloat("noise","sphere_radius1");
	sphere_centre2 = fin->ReadVec3f("noise","sphere_centre2");
	sphere_radius2 = fin->ReadFloat("noise","sphere_radius2");;
	d0 =0.5;
	d1 = 1.0;

	//ô˺ӵ״̬ʼ
	seed_particles(fin->ReadInteger("noise","particlenum"));

	//-------------------------------------------------------------

	//ʼ
	continuum = new Potential(fin->ReadVec3f("continuum","gridsize"),
							  fin->ReadFloat("continuum","gridlength"),
							  fin->ReadFloat("continuum","disvalue"));
	//ϵֵdiscomfortֵ
	continuum->resetAllGridData();

	//continuum
	continuum->SetGoal(fin->ReadVec3f("continuum","goalmin"), fin->ReadVec3f("continuum","goalmax"));
	continuum->SetSpeedRange(fin->ReadFloat("continuum","minspeed"),fin->ReadFloat("continuum","maxspeed"));
	continuum->SetDensityRange(fin->ReadFloat("continuum","mindensity"), fin->ReadFloat("continuum","maxdensity"));
	continuum->SetLamdaWeights(fin->ReadFloat("continuum","lamda"), 
							   fin->ReadFloat("continuum","w1"), 
							   fin->ReadFloat("continuum","w2"),
							   fin->ReadFloat("continuum","w3")); //lamdaٶȣʱ䣬Discomfort


	//0:PerlinNoise
	//1:RandomNoise
	//2:CurlNoise
	NoiseType = 2;

	//(octaves,frequency,amplitude,persistence,seed);
	//PNoiseX = new Perlin(4,1,5,1,45);
	//PNoiseY = new Perlin(4,1,5,1,46);
	//PNoiseZ = new Perlin(4,1,5,1,47);

	//PNoiseX = new Perlin(2,0.5,5,0.5,45);
	//PNoiseY = new Perlin(2,0.5,5,0.5,46);
	//PNoiseZ = new Perlin(2,0.5,5,0.5,47);

	//scale=0.4,gain=0.25
	//PNoiseX = new Perlin(4,2.5,2,1,45);
	//PNoiseY = new Perlin(4,2.5,2,1,46);
	//PNoiseZ = new Perlin(4,2.5,2,1,47);

	//scale=2,gain=0.5
	//PNoiseX = new Perlin(4,0.5,4,1,45);
	//PNoiseY = new Perlin(4,0.5,4,1,46);
	//PNoiseZ = new Perlin(4,0.5,4,1,47);

	//scale=5,gain=2
	PNoiseX = new Perlin(4,0.2,2,1,45);
	PNoiseY = new Perlin(4,0.2,2,1,46);
	PNoiseZ = new Perlin(4,0.2,2,1,47);

	//random
	//randomoffset=0.5;
	//randomoffset=1;
	randomoffset=2;
	count = 0;
}

void Simulator::AddPotentialDiscomfort(GLMmodel* model)
{
	//continuum->AddDiscomfort(model);
	//ܳ
	continuum->ComputeFaceData();
	continuum->FMM();
}
bool Simulator::seed_particles(int totalnum)
{
	Vec3f position(0, 0, 0);
	//һԲεĳʼ
	if(! strcmp(fin->ReadString("noise","shape"),"sphere"))
	{
		Vec3f centre = sphere_centre2;
		float radius1 = sphere_radius1, radius2 = sphere_radius2, radiusSqr1 = radius1 * radius1, radiusSqr2 = radius2 * radius2;

		for (int i = 0; i < totalnum; i ++)
		{
			while (1)
			{
				position = Vec3f(randhashf(seed ++, -radius2, radius2), randhashf(seed ++, -radius2, radius2), randhashf(seed ++, -radius2, radius2));
				float distance = mag2(position);
				if (distance < radiusSqr1 || distance > radiusSqr2)
					continue;
				position += centre;
				bool ingore = false;
				for (int j =0; j < getNumAgents(); j++)
					if (mag2(position - getAgentPosition(j))<=0.5)
					{
						ingore = true;
						break;
					}
				if (ingore)continue;
				addAgent(position);
				break;
			}
			
		}
		return true;
	}
	else 
	{
		//һĳʼ
		for (int i = 0; i < totalnum; i ++)
		{
			while (1)
			{
				position = Vec3f(randhashf(seed ++, initial_min[0], initial_max[0]), randhashf(seed ++, initial_min[1], initial_max[1]), randhashf(seed ++, initial_min[2], initial_max[2]));
				bool ignore = false;
				for (int j =0; j < getNumAgents(); j++)
					if (mag2(position - getAgentPosition(j))<=0.1)
					{
						ignore = true;
						break;
					}
					if (ignore) continue;
					addAgent(position);
					break;
			}

		}
		return true;
	}

	////bluenoise㷨ɵĳʼֲ
	//if(globalTime_ == 0){
	//	std::vector<Vec3f> x1;
	//	Particle p;
	//	p.node = 0;
	//	p.animTime = 0;
	//	bluenoise_sample(0.2f, Vec3f(initial_xmin, initial_ymin, initial_zmin), Vec3f(initial_xmax, initial_ymax, initial_zmax), x1);
	//	for (unsigned int i = 0; i < x1.size(); i ++)
	//	{
	//		p.x = x1[i];
	//		x.push_back(p);
	//	}
	//	return true;
	//}else
	//	return false; // no particles added after initial seeding at t=0

	////ʹwangtile㷨ɵĳʼֲ
	//if (t==0){
	//	LoadPoints(x,"ResultPoints.dat");
	//	return true;
	//}

	return false;

}

bool Simulator::reset_particles(int totalnum)
{

	Vec3f position(0, 0, 0);
	if(! strcmp(fin->ReadString("noise","shape"),"sphere"))
	{
		//һԲεĳʼ
		Vec3f centre = sphere_centre2;
		float radius1 = sphere_radius1, radius2 = sphere_radius2, radiusSqr1 = radius1 * radius1, radiusSqr2 = radius2 * radius2;

		for (int i = 0; i < totalnum; i ++)
		{
			while (1)
			{
				position = Vec3f(randhashf(seed ++, -radius2, radius2), randhashf(seed ++, -radius2, radius2), randhashf(seed ++, -radius2, radius2));
				float distance = mag2(position);
				if (distance < radiusSqr1 || distance > radiusSqr2)
					continue;
				position += centre;
				addAgent(position);
				break;
			}

		}
		return true;
	}
	else
	{
		//һĳʼ
		for (int i = 0; i < totalnum; i ++)
		{
			position = Vec3f(randhashf(seed ++, initial_min[0], initial_max[0]), randhashf(seed ++, initial_min[1], initial_max[1]), randhashf(seed ++, initial_min[2], initial_max[2]));
			addAgent(position);
		}
		return true;
	}



	//for (iterator i = particles.begin(); i != particles.end(); i++)
	//	(*i).updateDensity();

	//for(int x = 0; x < gridX; x ++)
	//	for (int y = 0; y < gridY; y ++)
	//		for(int z = 0; z < gridZ; z++)
	//			if(Particle::continuum.density[x][y][z] > 0) 
	//				Particle::continuum.averagev[x][y][z] = Particle::continuum.averagev[x][y][z] / Particle::continuum.density[x][y][z];

	return false;
}

Simulator::~Simulator()
{
	delete makenoise; 
	makenoise = 0x0; 

	delete continuum;
	continuum = 0x0;

	if (defaultAgent_ != NULL) {
		delete defaultAgent_;
		defaultAgent_ = NULL;
	}

	for (size_t i = 0; i < agents_.size(); ++i) {
		delete agents_[i];
		agents_[i] = NULL;
	}

	if (kdTree_ != NULL) {
		delete kdTree_;
		kdTree_ = NULL;
	}
}

float Simulator::getAgentAnimTime(size_t agentNo)
{
	return agents_[agentNo]->AnimTime;
}

size_t Simulator::getAgentNumAgentNeighbors(size_t agentNo) const
{
	return agents_[agentNo]->agentNeighbors_.size();
}

size_t Simulator::getAgentAgentNeighbor(size_t agentNo, size_t neighborNo) const
{
	return agents_[agentNo]->agentNeighbors_[neighborNo].second->id_;
}

size_t Simulator::getAgentNumORCAPlanes(size_t agentNo) const
{
	return agents_[agentNo]->orcaPlanes_.size();
}

const Plane& Simulator::getAgentORCAPlane(size_t agentNo, size_t planeNo) const
{
	return agents_[agentNo]->orcaPlanes_[planeNo];
}

void Simulator::removeAgent(size_t agentNo)
{
	delete agents_[agentNo];
	agents_[agentNo] = agents_.back();
	agents_.pop_back();
}

size_t Simulator::addAgent(const Vec3f& position)
{
	if (defaultAgent_ == NULL) {
		return RVO_ERROR;
	}

	Agent* agent = new Agent(this);

	agent->position_ = position;
	agent->maxNeighbors_ = defaultAgent_->maxNeighbors_;
	agent->maxSpeed_ = defaultAgent_->maxSpeed_;
	agent->neighborDist_ = defaultAgent_->neighborDist_;
	agent->radius_ = defaultAgent_->radius_;
	agent->timeHorizon_ = defaultAgent_->timeHorizon_;
	agent->velocity_ = defaultAgent_->velocity_;

	agent->id_ = agents_.size();

	agents_.push_back(agent);

	return agents_.size() - 1;
}

size_t Simulator::addAgent(const Vec3f& position, float neighborDist, size_t maxNeighbors, float timeHorizon, float radius, float maxSpeed, const Vec3f& velocity)
{
	Agent* agent = new Agent(this);

	agent->position_ = position;
	agent->maxNeighbors_ = maxNeighbors;
	agent->maxSpeed_ = maxSpeed;
	agent->neighborDist_ = neighborDist;
	agent->radius_ = radius;
	agent->timeHorizon_ = timeHorizon;
	agent->velocity_ = velocity;

	agent->id_ = agents_.size();

	agents_.push_back(agent);

	return agents_.size() - 1;
}

void Simulator::UpdateSpace()
{
	//ۼƼTotalDataеܶȺٶ section4.1е㷨
	for (int i = 0; i < static_cast<int>(agents_.size()); ++i) 
		continuum->addDensity(getAgentPosition(i), getAgentVelocity(i));

	//for(int x = 0; x < gridX; x ++)
	//	for (int y = 0; y < gridY; y ++)
	//		for(int z = 0; z < gridZ; z++)
	//			if(Particle::continuum.density[x][y][z] > 0) 
	//				continuum.averagev[x][y][z] = continuum.averagev[x][y][z] / continuum.density[x][y][z];

	//ܳ
	//continuum->ComputeFaceData();
	//continuum->FMM();

	//psi
	makenoise -> advance_time( timeStep_ );
}
void Simulator::doStep()
{
	kdTree_->buildAgentTree();

#ifdef _OPENMP
	#pragma omp parallel for
#endif
	for (int i = 0; i < static_cast<int>(agents_.size()); ++i) {
		agents_[i]->computeNeighbors();
		agents_[i]->computeNewVelocity();
	}

#ifdef _OPENMP
	#pragma omp parallel for
#endif
	for (int i = 0; i < static_cast<int>(agents_.size()); ++i) 
	{
		agents_[i]->update();
		//the codes below is used for collision detecting statistics.
		for (int j = 0; j < i; ++j) //Ѿλõ
			if ((dist(agents_[i]->position_ , agents_[j]->position_) < 2 * agents_[i]->radius_))
			{
				count++;
			}
		//----------------------------------------------------------
	}

	globalTime_ += timeStep_;
}

H3DNode Simulator::getAgentNode(size_t agentNo)
{
	return agents_[agentNo]->Node;
}
float Simulator::getAgentNeighborDist(size_t agentNo) const
{
	return agents_[agentNo]->neighborDist_;
}

const Vec3f& Simulator::getAgentPosition(size_t agentNo) const
{
	return agents_[agentNo]->position_;
}

const Vec3f& Simulator::getAgentPrefVelocity(size_t agentNo) const
{
	return agents_[agentNo]->prefVelocity_;
}

float Simulator::getAgentRadius(size_t agentNo) const
{
	return agents_[agentNo]->radius_;
}
 
const Vec3f& Simulator::getAgentVelocity(size_t agentNo) const
{
	return agents_[agentNo]->velocity_;
}

float Simulator::getGlobalTime() const
{
	return globalTime_;
}

size_t Simulator::getNumAgents() const
{
	return agents_.size();
}

float Simulator::getTimeStep() const
{
	return timeStep_;
}

void Simulator::setAgentAnimTime(size_t agentNo, float h3dAnimTime)
{
	agents_[agentNo]->AnimTime += h3dAnimTime;
}

void Simulator::setAgentDefaults(float neighborDist, size_t maxNeighbors, float timeHorizon, float radius, float maxSpeed, const Vec3f& velocity)
{
	if (defaultAgent_ == NULL) {
		defaultAgent_ = new Agent(this);
	}

	defaultAgent_->maxNeighbors_ = maxNeighbors;
	defaultAgent_->maxSpeed_ = maxSpeed;
	defaultAgent_->neighborDist_ = neighborDist;
	defaultAgent_->radius_ = radius;
	defaultAgent_->timeHorizon_ = timeHorizon;
	defaultAgent_->velocity_ = velocity;
}

void Simulator::setAgentMaxNeighbors(size_t agentNo, size_t maxNeighbors)
{
	agents_[agentNo]->maxNeighbors_ = maxNeighbors;
}

void Simulator::setAgentMaxSpeed(size_t agentNo, float maxSpeed)
{
	agents_[agentNo]->maxSpeed_ = maxSpeed;
}

void Simulator::setAgentNeighborDist(size_t agentNo, float neighborDist)
{
	agents_[agentNo]->neighborDist_ = neighborDist;
}

void Simulator::setAgentNode(size_t agentNo, H3DNode h3dnode)
{
	agents_[agentNo]->Node = h3dnode; 
}

void Simulator::setAgentPosition(size_t agentNo, const Vec3f& position)
{
	agents_[agentNo]->position_ = position;
}

void Simulator::setAgentPrefVelocity(size_t agentNo, const Vec3f& position)
{
	if (NoiseType == 2)
	{
		agents_[agentNo]->prefVelocity_ = continuum->GetVelocity(position) + makenoise->GetVelocity(position);
		if (continuum->isObstacle(position + agents_[agentNo]->prefVelocity_ * timeStep_ * 1.1))
			agents_[agentNo]->prefVelocity_ = continuum->GetVelocity(position);
	}
	else if (NoiseType == 0)
	{
		agents_[agentNo]->prefVelocity_ = continuum->GetVelocity(position) 
			+ Vec3f(PNoiseX->Get(position[0], position[1], position[2]) * 0.4, PNoiseY->Get(position[0], position[1], position[2]) * 0.4, PNoiseZ->Get(position[0], position[1], position[2]) * 0.4);
	}
	else
	{

		agents_[agentNo]->prefVelocity_ = continuum->GetVelocity(position) 
			+ Vec3f(rand() % int(randomoffset*100*2) / 100.0 - randomoffset, rand() % int(randomoffset*100*2) / 100.0 - randomoffset, rand() % int(randomoffset*100*2) / 100.0 - randomoffset);
	}

}

void Simulator::setAgentRadius(size_t agentNo, float radius)
{
	agents_[agentNo]->radius_ = radius;
}


void Simulator::setAgentVelocity(size_t agentNo, const Vec3f& velocity)
{
	agents_[agentNo]->velocity_ = velocity;
}

void Simulator::setTimeStep(float timeStep)
{
	timeStep_ = timeStep;
}
}
