//app.h

#pragma once

#ifndef _app_H_
#define _app_H_

#include "makenoise.h"
#include "Potential.h"
#include <string>
#include "vec.h"
#include "Simulator.h"

class Application
{
public:
	Application( const std::string &appPath );
	
	void setKeyState( int key, bool state ) { _prevKeys[key] = _keys[key]; _keys[key] = state; }

	const char *getTitle() { return "Swarm"; }
	
	bool init();
	void mainLoop( float fps );
	void release();
	void resize( int width, int height );

	void keyStateHandler();
	void mouseMoveEvent( float dX, float dY );
	void update(float fps);

private:
	void keyHandler();
	void init_day();
	void init_lamp();
	void reset();
	GLMmodel* addObj(const char* filename);
	void translateModel(GLMmodel* model, Vec3f p);
	void Muti( float matrix[4][4], float matrix1[4][4], float matrix2[4][4] );
	void Rotate( float costheta, float sintheta, float nx, float ny, float nz, float matrix[4][4] );

private:
	bool         _keys[320], _prevKeys[320];

	float        _x, _y, _z, _rx, _ry, _rz;  // Viewer position and orientation
	float        _velocity;  // Velocity for movement
	float        _curFPS;
	float		 display_timestep, simulation_timestep;  


	int          _statMode;
	int          _freezeMode;
	bool         _debugViewMode, _wireframeMode;
	int			 junk;

	static const int tracelength = 1;
	
	RVO::Simulator* sim;

	// Engine objects
	H3DRes       _fontMatRes, _panelMatRes;
	H3DRes       _logoMatRes, _forwardPipeRes, _deferredPipeRes;
	H3DNode      _cam;

	std::string  _contentDir;

	//---------------------------------------
	int gridX;
	int gridY;
	int gridZ;
	
	//OBJ任 ˵readme.txt
	Vec3f positionA, positionB, OBJtranslate, moveworld;
	float OBJscale;

	int frame;

	//---------------------------------------

};


#endif // _app_H_


