//Potential.cpp 
#include "Potential.h"
#include "interp.h"
#include <cmath>
//extern FILE *fin, *fout;

Potential::Potential(Vec3f size, float _GridLength, float _DisValue) 
:GridLength(_GridLength), DisValue(_DisValue) 
{
	gridX = floor(size[0] / GridLength + 0.5);
	gridY = floor(size[1] / GridLength + 0.5);
	gridZ = floor(size[2] / GridLength + 0.5);
	tmpgridX = gridX - 2; 
	tmpgridY = gridY - 2;
	tmpgridZ = gridZ - 2;
	totalGrids = (tmpgridX - 2) * (tmpgridY - 2) * (tmpgridZ - 2);//
	Candidate.resize(totalGrids);
}

void Potential::SetGoal(Vec3f goalmin, Vec3f goalmax)
{
	m_GoalXa = goalmin[0] / GridLength;
	m_GoalXb = goalmax[0] / GridLength;
	m_GoalYa = goalmin[1] / GridLength;
	m_GoalYb = goalmax[1] / GridLength;
	m_GoalZa = goalmin[2] / GridLength;
	m_GoalZb = goalmax[2] / GridLength;
}

void Potential::SetSpeedRange(float _minSpeed, float _maxSpeed)
{
	maxSpeed = _maxSpeed;
	minSpeed = _minSpeed;
}

void Potential::SetDensityRange(float _minDensity, float _maxDensity)
{
	maxDensity = _maxDensity;
	minDensity = _minDensity;
}

void Potential::SetLamdaWeights(float _lamda, float _w1, float _w2, float _w3)
{
	lamda = _lamda;
	w1 = _w1;
	w2 = _w2;
	w3 = _w3;
}

void Potential::resetAllGridData(void)
{
	for(int x = 0; x < gridX; x ++)
		for (int y = 0; y < gridY; y ++)
			for(int z = 0; z < gridZ; z ++)
			{
				density[x][y][z] = 0;
				averagev[x][y][z] = Vec3f(0,0,0);
				m_gridData[x][y][z].discomfort = 0;
			}

}

void Potential::ComputeFaceData(void)
{
	int x, y, z;
	float speedT = minSpeed, speedV;
	for(x = 0; x < gridX; x ++)
		for(y = 0; y < gridY; y ++)
			for (z = 0; z < gridZ; z ++)
			{
				speedV = mag(get_DensityGradient(Vec3f(x, y, z) / GridLength));
				if (x + 1 < gridX)
				{
					//speedV =  averagev[x + 1][y][z][0];
					if(speedV > speedT) speedV = speedT;
					if(density[x + 1][y][z] <= minDensity)
					{
						m_gridData[x][y][z].faceData[0].speed = speedT;
					}
					else if(density[x + 1][y][z] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[0].speed = speedV;
					}
					else
					{
						m_gridData[x][y][z].faceData[0].speed = speedT + (density[x + 1][y][z] - minDensity)
							/(maxDensity - minDensity) * (speedV-speedT);
					}
				}
				if (y + 1 < gridY)
				{
					//speedV =  abs(averagev[x][y + 1][z][1]);
					if(speedV > speedT) speedV = speedT;
					if(density[x][y + 1][z] <= minDensity)
					{
						m_gridData[x][y][z].faceData[1].speed = speedT;
					}
					else if(density[x][y + 1][z] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[1].speed = speedV;
					}
					else
					{
						m_gridData[x][y][z].faceData[1].speed = speedT + (density[x][y + 1][z] - minDensity)
							/(maxDensity - minDensity) * (speedV-speedT);
					}
				}
				if (z + 1 < gridZ)
				{
					//speedV =  abs(averagev[x][y][z + 1][2]);
					if(speedV > speedT) speedV = speedT;
					if(density[x][y][z + 1] <= minDensity)
					{
						m_gridData[x][y][z].faceData[2].speed = speedT;
					}
					else if(density[x][y][z + 1] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[2].speed = speedV;
					}
					else
					{
						m_gridData[x][y][z].faceData[2].speed = speedT + (density[x][y][z + 1] - minDensity)
							/(maxDensity - minDensity) * (speedV-speedT);
					}
				}
				if (x - 1 >= 0)
				{
					//speedV =  -averagev[x - 1][y][z][0];
					if(speedV < 0) speedV = 0;
					if(speedV > speedT) speedV = speedT;
					if(density[x - 1][y][z] <= minDensity)
					{
						m_gridData[x][y][z].faceData[3].speed = speedT;
					}
					else if(density[x - 1][y][z] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[3].speed = speedV;
					}
					else
					{ 
						m_gridData[x][y][z].faceData[3].speed = speedT + (density[x - 1][y][z] - minDensity) 
							/ (maxDensity - minDensity) * (speedV - speedT);
					}
				}
				if (y - 1 >= 0)
				{
					//speedV =  -averagev[x][y - 1][z][1];
					if(speedV < 0) speedV = 0;
					if(speedV > speedT) speedV = speedT;
					if(density[x][y - 1][z] <= minDensity)
					{
						m_gridData[x][y][z].faceData[4].speed = speedT;
					}
					else if(density[x][y - 1][z] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[4].speed = speedV;
					}
					else
					{ 
						m_gridData[x][y][z].faceData[4].speed = speedT + (density[x][y - 1][z] - minDensity) 
							/ (maxDensity - minDensity) * (speedV - speedT);
					}
				}
				if(z - 1 >= 0)
				{
					//speedV =  -averagev[x][y][z - 1][2];
					if(speedV < 0) speedV = 0;
					if(speedV > speedT) speedV = speedT;
					if(density[x][y][z - 1] <= minDensity)
					{
						m_gridData[x][y][z].faceData[5].speed = speedT;
					}
					else if(density[x][y][z - 1] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[5].speed = speedV;
					}
					else
					{ 
						m_gridData[x][y][z].faceData[5].speed = speedT + (density[x][y][z - 1] - minDensity) 
							/ (maxDensity - minDensity) * (speedV - speedT);
					}
				}
			}

	for(x = 0; x < gridX; x ++)
		for (y = 0; y < gridY; y ++) 
			for( z = 0; z < gridZ; z ++)
			{
				if(x + 1 < gridX)
				{	
					if(m_gridData[x][y][z].faceData[0].speed <= 0) 
						m_gridData[x][y][z].faceData[0].speed = MININT;
					m_gridData[x][y][z].faceData[0].cost = (w1 * m_gridData[x][y][z].faceData[0].speed + w2 +
						w3 * m_gridData[x + 1][y][z].discomfort) / m_gridData[x][y][z].faceData[0].speed;
				}
				if(y + 1 < gridY)
				{	
					if(m_gridData[x][y][z].faceData[1].speed <= 0) 
						m_gridData[x][y][z].faceData[1].speed = MININT;
					m_gridData[x][y][z].faceData[1].cost = (w1 * m_gridData[x][y][z].faceData[1].speed + w2 +
						w3 * m_gridData[x][y + 1][z].discomfort) / m_gridData[x][y][z].faceData[1].speed;
				}
				if(z + 1 < gridZ)
				{	
					if(m_gridData[x][y][z].faceData[2].speed <= 0) 
						m_gridData[x][y][z].faceData[2].speed = MININT;
					m_gridData[x][y][z].faceData[2].cost = (w1 * m_gridData[x][y][z].faceData[2].speed + w2 +
						w3 * m_gridData[x][y][z + 1].discomfort) / m_gridData[x][y][z].faceData[2].speed;
				}

				if(x - 1 >= 0)
				{
					if(m_gridData[x][y][z].faceData[3].speed <= 0) 
						m_gridData[x][y][z].faceData[3].speed = MININT;
					m_gridData[x][y][z].faceData[3].cost = (w1 * m_gridData[x][y][z].faceData[3].speed + w2 + 
						w3 * m_gridData[x - 1][y][z].discomfort) / m_gridData[x][y][z].faceData[3].speed;
				}
				if(y - 1 >= 0)
				{
					if(m_gridData[x][y][z].faceData[4].speed <= 0) 
						m_gridData[x][y][z].faceData[4].speed = MININT;
					m_gridData[x][y][z].faceData[4].cost = (w1 * m_gridData[x][y][z].faceData[4].speed + w2 + 
						w3 * m_gridData[x][y - 1][z].discomfort) / m_gridData[x][y][z].faceData[4].speed;
				}
				if(z - 1 >= 0)
				{
					if(m_gridData[x][y][z].faceData[5].speed <= 0) 
						m_gridData[x][y][z].faceData[5].speed = MININT;
					m_gridData[x][y][z].faceData[5].cost = (w1 * m_gridData[x][y][z].faceData[5].speed + w2 + 
						w3 * m_gridData[x][y][z - 1].discomfort) / m_gridData[x][y][z].faceData[5].speed;
				}
			}
}


void Potential::AddToHeap(int x, int y, int z)
{
	int index = GI(x, y, z);
	int x1, y1, z1, x2, y2, z2;
	Candidate[heapSize] = index;
	m_gridData[x][y][z].HeapPosition = heapSize;

	int j, i = heapSize;
	for(i; i > 0; i = j) 
	{
		j = (i-1)/2;
		GIJK(Candidate[i], x1, y1, z1);
		GIJK(Candidate[j], x2, y2, z2);
		if (m_gridData[x1][y1][z1].potential < m_gridData[x2][y2][z2].potential )
		{
			Candidate[i] = Candidate[j];
			m_gridData[x2][y2][z2].HeapPosition = i;
			Candidate[j] = index;
			m_gridData[x][y][z].HeapPosition = j;
		}
		else
			break;
	}

	heapSize ++;
}

void Potential::UpdateHeap(int x, int y, int z) 
{
	int x1, y1, z1, x2, y2, z2;
	int index = GI(x, y, z);
	int j, i = m_gridData[x][y][z].HeapPosition;
	for(i; i > 0; i = j) 
	{
		j = (i-1)/2;
		GIJK(Candidate[i], x1, y1, z1);
		GIJK(Candidate[j], x2, y2, z2);
		if (m_gridData[x1][y1][z1].potential < m_gridData[x2][y2][z2].potential ) 
		{
			Candidate[i] = Candidate[j];
			m_gridData[x2][y2][z2].HeapPosition = i;
			Candidate[j] = index;
			m_gridData[x][y][z].HeapPosition = j;
		}
		else
			break;
	}
}

int Potential::PopHeap() 
{
	if(heapSize == 0)
		return -1;
	int j, index = Candidate[0];
	int x, y ,z;
	GIJK(index, x, y ,z);
	heapSize --;

	Candidate[0] = Candidate[heapSize];
	GIJK(Candidate[heapSize], x, y ,z);
	m_gridData[x][y][z].HeapPosition = 0;
	for(int i = 0; i < (heapSize-1); i = j) 
	{
		int lc = 2*i+1;
		int rc = 2*i+2;
		GIJK(Candidate[i], x, y, z);
		float current = m_gridData[x][y][z].potential;
		float lv, rv;
		if(lc < heapSize) 
		{
			GIJK(Candidate[lc], x, y, z);
			lv = m_gridData[x][y][z].potential;
			if(rc < heapSize) 
			{
				GIJK(Candidate[rc], x, y, z);
				rv = m_gridData[x][y][z].potential;
				if(lv > rv) {
					lc = rc;
					lv = rv;
				}
			}
			if(current > lv) {
				Candidate[i] = Candidate[lc];
				GIJK(Candidate[i], x, y, z);
				m_gridData[x][y][z].HeapPosition = i;
				Candidate[lc] = Candidate[heapSize];
				GIJK(Candidate[heapSize], x, y, z);
				m_gridData[x][y][z].HeapPosition = lc;
				j = lc;
			}
			else
				break;
		}
		else
			break;
	}
	return index;
}

//processڵtargetΪCANDIDATE״̬CandidateУPotential
void Potential::process(int x, int y, int z)
{
	//ýڵѾõСֵ򲻱ؼ
	if (m_gridData[x][y][z].FMMFlag == KNOWN)
		return;

	CalcPotential(x, y, z);

	if (m_gridData[x][y][z].FMMFlag == CANDIDATE)
	{
		UpdateHeap(x, y, z);
	}
	else
	{
		m_gridData[x][y][z].FMMFlag = CANDIDATE;
		AddToHeap(x, y, z);
	}
}
void Potential::FMM()
{
	int x, y, z;

	//ʼ еPotentialΪ״̬ΪUNKOWN
	for(x = 0; x < gridX; x ++)
		for (y = 0; y < gridY; y ++)
			for(z = 0; z < gridZ; z ++)
			{
				//ʼ״̬ΪɴΪֵ
				m_gridData[x][y][z].FMMFlag = UNKNOWN;
				m_gridData[x][y][z].potential = MAXINT;
			}

	heapSize = 0;
	//ĿPotentialΪ0״̬ΪKNOWN.
	for(x = m_GoalXa; x <= m_GoalXb; x ++)
		for (y = m_GoalYa; y <= m_GoalYb; y ++)
			for(z = m_GoalZa; z <= m_GoalZb; z ++)
			{
				m_gridData[x][y][z].FMMFlag = KNOWN;
				m_gridData[x][y][z].potential = 0;
			}
	//ʼCandidateѡ
	for(x = m_GoalXa; x <= m_GoalXb; x ++)
		for (y = m_GoalYa; y <= m_GoalYb; y ++)
			for(z = m_GoalZa; z <= m_GoalZb; z ++)
			{
				if (x + 1 < tmpgridX) process(x + 1, y, z);
				if (y + 1 < tmpgridY) process(x, y + 1, z);
				if (z + 1 < tmpgridZ) process(x, y, z + 1);

				if (x - 1 >= 2) process(x - 1, y, z);
				if (y - 1 >= 2) process(x, y - 1, z);
				if (z - 1 >= 2) process(x, y, z - 1);
			}

	/*ѵĴСheapSizeΪ0ʱѭȰCandidatePotentialֵСĽڵΪKNOWNýڵھӣ㣩
	  CandidateвPotentialʽ11*/
	int tmp;
	while ((tmp = PopHeap()) != -1)
	{
		GIJK(tmp, x, y, z);

		m_gridData[x][y][z].FMMFlag = KNOWN;
		if (x + 1 < tmpgridX) process(x + 1, y, z);
		if (y + 1 < tmpgridY) process(x, y + 1, z);
		if (z + 1 < tmpgridZ) process(x, y, z + 1);

		if (x - 1 >= 2) process(x - 1, y, z);
		if (y - 1 >= 2) process(x, y - 1, z);
		if (z - 1 >= 2) process(x, y, z - 1);
	}

	return;
}
void Potential::solve(float &x, float a, float b, float c, float d)
{
	float A = d + b;
	float B = - 2 * a * d - 2 * c * b;
	float C = a * a * d + c * c * b - b * d;
	float D = B * B - 4 * A * C;
	if(D < 0) D = 0;
	x = (- B + sqrt(D)) / (2.0 * A);
}
void Potential::CalcPotential(int x, int y, int z)
{
	float totalmin = MAXINT;
	bool minx = false;
	float a, b = 0, c, d = 0, e ,f = 0;
	if (x < gridX - 1 && m_gridData[x + 1][y][z].FMMFlag == KNOWN && m_gridData[x + 1][y][z].potential + m_gridData[x][y][z].faceData[0].cost < totalmin)
	{
		a = m_gridData[x + 1][y][z].potential;
		b = m_gridData[x][y][z].faceData[0].cost;
		totalmin = a + b;
		minx = true; 
	}
	if (x > 0 && m_gridData[x - 1][y][z].FMMFlag == KNOWN && m_gridData[x - 1][y][z].potential + m_gridData[x][y][z].faceData[3].cost < totalmin)
	{
		a = m_gridData[x - 1][y][z].potential;
		b = m_gridData[x][y][z].faceData[3].cost;
		minx = true;
	}

	totalmin = MAXINT;
	bool miny = false;
	if (y < gridY - 1 && m_gridData[x][y + 1][z].FMMFlag == KNOWN && m_gridData[x][y + 1][z].potential + m_gridData[x][y][z].faceData[1].cost < totalmin)
	{
		c = m_gridData[x][y + 1][z].potential;
		d = m_gridData[x][y][z].faceData[1].cost;
		totalmin = c + d;
		miny = true;
	}
	if (y > 0 && m_gridData[x][y - 1][z].FMMFlag == KNOWN && m_gridData[x][y - 1][z].potential + m_gridData[x][y][z].faceData[4].cost < totalmin)
	{
		c = m_gridData[x][y - 1][z].potential;
		d = m_gridData[x][y][z].faceData[4].cost;
		miny = true;
	}

	totalmin = MAXINT;
	bool minz = false;
	if (z < gridZ - 1 && m_gridData[x][y][z + 1].FMMFlag == KNOWN && m_gridData[x][y][z + 1].potential + m_gridData[x][y][z].faceData[2].cost < totalmin)
	{
		e = m_gridData[x][y][z + 1].potential;
		f = m_gridData[x][y][z].faceData[2].cost;
		totalmin = e + f;
		minz = true;
	}
	if (z > 0 && m_gridData[x][y][z - 1].FMMFlag == KNOWN && m_gridData[x][y][z - 1].potential + m_gridData[x][y][z].faceData[5].cost < totalmin)
	{
		e = m_gridData[x][y][z - 1].potential;
		f = m_gridData[x][y][z].faceData[5].cost;
		minz = true;
	}

	//㹫ʽ11.㷨Ҫ㲻֪дôuglyĳȫelse if
	if (!minx && !miny)
		m_gridData[x][y][z].potential = e + f;
	else if (!minx && !minz)
		m_gridData[x][y][z].potential = c + d;
	else if (!miny && !minz)
		m_gridData[x][y][z].potential = a + b;
	else 
	{
		//ƽü㡣
		b = b * b;
		d = d * d;
		f = f * f;

		if (!minx)
			solve(m_gridData[x][y][z].potential, c, d, e, f);
		else if (!miny)
			solve(m_gridData[x][y][z].potential, a, b, e, f);
		else if (!minz)
			solve(m_gridData[x][y][z].potential, a, b, c, d);
		else
		{
			//ߵһ˵·̫ˡ
			float A = d * f + b * f + b * d;
			float B = - 2 * d * f * a - 2 * b * f * c - 2 * b * d * e;
			float C = a * a * d * f + c * c * b * f + e * e * b * d - b * d * f;
			float D = B * B - 4 * A * C;
			if (D < 0) D = 0;
			m_gridData[x][y][z].potential = (- B + sqrt(D)) / (2.0 * A);
		}

		//if (!minx && !miny && !minz)
		//{
		//	minx = 0;
		//}
	}
}
void Potential::addDensity(Vec3f position, Vec3f velocity)
{
	float x = position[0] / GridLength;
	float y = position[1] / GridLength;
	float z = position[2] / GridLength;
	int i = floor(x + 0.5);
	int j = floor(y + 0.5);
	int k = floor(z + 0.5);
	//Ӧsection 4.1㷨
	if (i > gridX || j > gridY || k > gridZ || i < 0 || j < 0 || k < 0) return;

	float deltax = x - i + 0.5 * GridLength;
	float deltay = y - j + 0.5 * GridLength;
	float deltaz = z - k + 0.5 * GridLength;
	float tmp;
	if (i > 0 && j > 0 && k < gridZ)  //density A
	{
		tmp = min(1 - deltax, 1 - deltay, deltaz);
		tmp = pow(tmp, lamda);
		density[i - 1][j - 1][k] += tmp;
		averagev[i - 1][j - 1][k] += velocity * tmp;
	}
	if (i < gridX && j > 0 && k < gridZ) //density B
	{
		tmp = min(deltax, 1 - deltay, deltaz);
		tmp = pow(tmp, lamda);
		density[i][j - 1][k] += tmp;
		averagev[i][j - 1][k] += velocity * tmp;
	}
	if (i < gridX && j < gridY && k < gridZ) //density C
	{
		tmp = min(deltax, deltay, deltaz);
		tmp = pow(tmp, lamda);
		density[i][j][k] += tmp;
		averagev[i][j][k] += velocity * tmp;
	}
	if (i > 0 && j < gridY && k < gridZ) //density D
	{
		tmp = min(1 - deltax, deltay, deltaz);
		tmp = pow(tmp, lamda);
		density[i - 1][j][k] += tmp;
		averagev[i - 1][j][k] += velocity * tmp;
	}

	if (i > 0 && j > 0 && k > 0)  //density E
	{
		tmp = min(1 - deltax, 1 - deltay, 1 - deltaz);
		tmp = pow(tmp, lamda);
		density[i - 1][j - 1][k - 1] += tmp;
		averagev[i - 1][j - 1][k - 1] += velocity * tmp;
	}
	if (i < gridX && j > 0 && k > 0) //density F
	{
		tmp = min(deltax, 1 - deltay, 1 - deltaz);
		tmp = pow(tmp, lamda);
		density[i][j - 1][k - 1] += tmp;
		averagev[i][j - 1][k - 1] += velocity * tmp;
	}
	if (i < gridX && j < gridY && k > 0) //density G
	{
		tmp = min(deltax, deltay, 1 - deltaz);
		tmp = pow(tmp, lamda);
		density[i][j][k - 1] += tmp;
		averagev[i][j][k - 1] += velocity * tmp;
	}
	if (i > 0 && j < gridY && k > 0) //density H
	{
		tmp = min(1 - deltax, deltay, 1 - deltaz);
		tmp = pow(tmp, lamda);
		density[i - 1][j][k - 1] += tmp;
		averagev[i - 1][j][k - 1] += velocity * tmp;
	}
}
Vec3f Potential::GetVelocity(Vec3f position)
{
	//ֵgradient
	Vec3f gradient = get_PotentialGradient(position / GridLength);
	gradient = normalized(-gradient);

	//i,j,kʾposition
	int i = floor(position[0] / GridLength);
	int j = floor(position[1] / GridLength);
	int k = floor(position[2] / GridLength);
	if (m_gridData[i][j][k].potential == 0) 
		return(Vec3f(0, 0, 0));
	//speed
	//x,y,zʾpositiongradientλ
	int x = floor(position[0] / GridLength + gradient[0] / 2 + 0.5);
	int y = floor(position[1] / GridLength + gradient[1] / 2 + 0.5);
	int z = floor(position[2] / GridLength + gradient[2] / 2 + 0.5);
	float speed = m_gridData[i][j][k].faceData[(-3) * (x - i) + 3].speed * abs(gradient[0]) + 
		m_gridData[i][j][k].faceData[(-3) * (y - j) + 4].speed * abs(gradient[1]) +
		m_gridData[i][j][k].faceData[(-3) * (z - k) + 5].speed * abs(gradient[2]);
	Vec3f velocity = gradient * speed;
	return velocity;
}

bool Potential::isObstacle(Vec3f position)
{
	//i,j,kʾposition
	int i = floor(position[0] / GridLength);
	int j = floor(position[1] / GridLength);
	int k = floor(position[2] / GridLength);
	if (m_gridData[i][j][k].discomfort > 0) 
		return true;
	else 
		return false;
}

Vec3f Potential::get_PotentialGradient(Vec3f position)
{
	int i,j,k;
	float fx,fy,fz;
	Vec3f gradient;

	get_barycentric(position[0], i, fx, 0, gridX);
	get_barycentric(position[1], j, fy, 0, gridY);
	get_barycentric(position[2], k, fz, 0, gridZ);
	//dx
	float v000 = (m_gridData[i+1][j][k].potential - m_gridData[i-1][j][k].potential) / GridLength * 0.5f;
	float v001 = (m_gridData[i+1][j][k+1].potential - m_gridData[i-1][j][k+1].potential) / GridLength * 0.5f;
	float v010 = (m_gridData[i+1][j+1][k].potential - m_gridData[i-1][j+1][k].potential) / GridLength*0.5f;
	float v011 = (m_gridData[i+1][j+1][k+1].potential - m_gridData[i-1][j+1][k+1].potential) / GridLength * 0.5f;
	float v100 = (m_gridData[i+2][j][k].potential - m_gridData[i][j][k].potential) / GridLength * 0.5f;
	float v101 = (m_gridData[i+2][j][k+1].potential-m_gridData[i][j][k+1].potential) / GridLength * 0.5f;
	float v110 = (m_gridData[i+2][j+1][k].potential-m_gridData[i][j+1][k].potential) / GridLength * 0.5f;
	float v111 = (m_gridData[i+2][j+1][k+1].potential - m_gridData[i][j+1][k+1].potential) / GridLength * 0.5f;
	gradient[0]=trilerp(v000, v100, v010, v110, v001, v101, v011, v111, fx, fy, fz);
	//dy
	v000 = (m_gridData[i][j+1][k].potential-m_gridData[i][j-1][k].potential)/GridLength*0.5f;
	v001 = (m_gridData[i][j+1][k+1].potential-m_gridData[i][j-1][k+1].potential)/GridLength*0.5f;
	v010 = (m_gridData[i][j+2][k].potential-m_gridData[i][j][k].potential)/GridLength*0.5f;
	v011 = (m_gridData[i][j+2][k+1].potential-m_gridData[i][j][k+1].potential)/GridLength*0.5f;
	v100 = (m_gridData[i+1][j+1][k].potential-m_gridData[i+1][j-1][k].potential)/GridLength*0.5f;
	v101 = (m_gridData[i+1][j+1][k+1].potential-m_gridData[i+1][j-1][k+1].potential)/GridLength*0.5f;
	v110 = (m_gridData[i+1][j+2][k].potential-m_gridData[i+1][j][k].potential)/GridLength*0.5f;
	v111 = (m_gridData[i+1][j+2][k+1].potential-m_gridData[i+1][j][k+1].potential)/GridLength*0.5f;
	gradient[1]=trilerp(v000, v100,v010, v110, v001, v101,v011, v111,fx, fy, fz);
	//dz
	v000 = (m_gridData[i][j][k+1].potential-m_gridData[i][j][k-1].potential)/GridLength*0.5f;
	v001 = (m_gridData[i][j][k+2].potential-m_gridData[i][j][k].potential)/GridLength*0.5f;
	v010 = (m_gridData[i][j+1][k+1].potential-m_gridData[i][j+1][k-1].potential)/GridLength*0.5f;
	v011 = (m_gridData[i][j+1][k+2].potential-m_gridData[i][j+1][k].potential)/GridLength*0.5f;
	v100 = (m_gridData[i+1][j][k+1].potential-m_gridData[i+1][j][k-1].potential)/GridLength*0.5f;
	v101 = (m_gridData[i+1][j][k+2].potential-m_gridData[i+1][j][k].potential)/GridLength*0.5f;
	v110 = (m_gridData[i+1][j+1][k+1].potential-m_gridData[i+1][j+1][k-1].potential)/GridLength*0.5f;
	v111 = (m_gridData[i+1][j+1][k+2].potential-m_gridData[i+1][j+1][k].potential)/GridLength*0.5f;
	gradient[2]=trilerp(v000, v100,v010, v110, v001, v101,v011, v111,fx, fy, fz);

	return gradient;
}

Vec3f Potential::get_DensityGradient(Vec3f position)
{
	int i,j,k;
	float fx,fy,fz;
	Vec3f gradient;

	get_barycentric(position[0], i, fx, 0, gridX);
	get_barycentric(position[1], j, fy, 0, gridY);
	get_barycentric(position[2], k, fz, 0, gridZ);
	//dx
	float v000 = (density[i+1][j][k] - density[i-1][j][k])* 0.5f;
	float v001 = (density[i+1][j][k+1] - density[i-1][j][k+1])* 0.5f;
	float v010 = (density[i+1][j+1][k] - density[i-1][j+1][k])*0.5f;
	float v011 = (density[i+1][j+1][k+1] - density[i-1][j+1][k+1])* 0.5f;
	float v100 = (density[i+2][j][k] - density[i][j][k])* 0.5f;
	float v101 = (density[i+2][j][k+1]-density[i][j][k+1])* 0.5f;
	float v110 = (density[i+2][j+1][k]-density[i][j+1][k])* 0.5f;
	float v111 = (density[i+2][j+1][k+1] - density[i][j+1][k+1])* 0.5f;
	gradient[0]=trilerp(v000, v100, v010, v110, v001, v101, v011, v111, fx, fy, fz);
	//dy
	v000 = (density[i][j+1][k]-density[i][j-1][k])*0.5f;
	v001 = (density[i][j+1][k+1]-density[i][j-1][k+1])*0.5f;
	v010 = (density[i][j+2][k]-density[i][j][k])*0.5f;
	v011 = (density[i][j+2][k+1]-density[i][j][k+1])*0.5f;
	v100 = (density[i+1][j+1][k]-density[i+1][j-1][k])*0.5f;
	v101 = (density[i+1][j+1][k+1]-density[i+1][j-1][k+1])*0.5f;
	v110 = (density[i+1][j+2][k]-density[i+1][j][k])*0.5f;
	v111 = (density[i+1][j+2][k+1]-density[i+1][j][k+1])*0.5f;
	gradient[1]=trilerp(v000, v100,v010, v110, v001, v101,v011, v111,fx, fy, fz);
	//dz
	v000 = (density[i][j][k+1]-density[i][j][k-1])*0.5f;
	v001 = (density[i][j][k+2]-density[i][j][k])*0.5f;
	v010 = (density[i][j+1][k+1]-density[i][j+1][k-1])*0.5f;
	v011 = (density[i][j+1][k+2]-density[i][j+1][k])*0.5f;
	v100 = (density[i+1][j][k+1]-density[i+1][j][k-1])*0.5f;
	v101 = (density[i+1][j][k+2]-density[i+1][j][k])*0.5f;
	v110 = (density[i+1][j+1][k+1]-density[i+1][j+1][k-1])*0.5f;
	v111 = (density[i+1][j+1][k+2]-density[i+1][j+1][k])*0.5f;
	gradient[2]=trilerp(v000, v100,v010, v110, v001, v101,v011, v111,fx, fy, fz);

	return gradient;
}

bool Potential::AddDiscomfort(GLMmodel* model)
{
	//ʼdistance
	int i, j, k;
	for (i = 0 ;i < gridX; i ++)
		for (j = 0; j < gridY; j ++)
			for (k = 0; k < gridZ; k ++)
				distance[i][j][k] = MAXINT;

	//ߵķ
	Vertex_and_edge_Normals(model,edge_normals);

	int near_i,near_j,near_k;
	int i_min,i_max,j_min,j_max,k_min,k_max;
	Vec3f point_a,point_b,point_c;
	int index_a,index_b,index_c;
	for (i = 0; i < model->numtriangles; i ++)
	{
		index_a=model->triangles[i].vindices[0];
		index_b=model->triangles[i].vindices[1];
		index_c=model->triangles[i].vindices[2];

		point_a[0]=model->vertices[3*index_a];
		point_a[1]=model->vertices[3*index_a+1];
		point_a[2]=model->vertices[3*index_a+2];

		point_b[0]=model->vertices[3*index_b];
		point_b[1]=model->vertices[3*index_b+1];
		point_b[2]=model->vertices[3*index_b+2];

		point_c[0]=model->vertices[3*index_c];
		point_c[1]=model->vertices[3*index_c+1];
		point_c[2]=model->vertices[3*index_c+2];

		i_max=max((point_a[0])/GridLength,(point_b[0])/GridLength,(point_c[0])/GridLength);
		i_min=min((point_a[0])/GridLength,(point_b[0])/GridLength,(point_c[0])/GridLength);

		j_max=max((point_a[1])/GridLength,(point_b[1])/GridLength,(point_c[1])/GridLength);
		j_min=min((point_a[1])/GridLength,(point_b[1])/GridLength,(point_c[1])/GridLength);

		k_max=max((point_a[2])/GridLength,(point_b[2])/GridLength,(point_c[2])/GridLength);
		k_min=min((point_a[2])/GridLength,(point_b[2])/GridLength,(point_c[2])/GridLength);

		for (near_k=clamp(k_min-3,0,gridZ - 1);near_k<clamp(k_max+3,0,gridZ - 1);near_k++)
			for (near_j=clamp(j_min-3,0,gridY - 1);near_j<clamp(j_max+3,0,gridY - 1);near_j++)
				for (near_i=clamp(i_min-3,0,gridX - 1);near_i<clamp(i_max+3,0,gridX - 1);near_i++)
				{
					float dis;
					Vec3f grid_point((near_i + 0.5) * GridLength, (near_j + 0.5) * GridLength, (near_k + 0.5) * GridLength);
					dis = ComputeDistance(model, i, index_a, index_b, index_c, point_a, point_b, point_c, grid_point);
					if (fabs(dis) < fabs(distance[near_i][near_j][near_k]))
						distance[near_i][near_j][near_k] = dis;
				}
	}

	////FLOOD FILL
	//list *head, *tail, *creat;
	//head = new list;
	//head->x = gridX - 1;
	//head->y = gridY - 1;
	//head->z = gridZ - 1;
	//head->next = NULL;
	//distance[head->x][head->y][head->z] = -MAXINT;
	//tail = head; //βָ

	//while (head != NULL) //лԪʱ
	//{
	//	i = head->x; j = head->y; k = head->z;
	//	if (i > 0 && distance[i - 1][j][k] == MAXINT) //ھδֵ״̬
	//	{
	//		distance[i - 1][j][k] = -MAXINT;
	//		creat = new list;
	//		creat->x = i - 1;
	//		creat->y = j;
	//		creat->z = k;
	//		creat->next = NULL;
	//		tail->next = creat; //βԪ
	//		tail = tail->next;  //βƶһ
	//	}
	//	if ((i < gridX - 1) && distance[i + 1][j][k] == MAXINT)
	//	{
	//		distance[i + 1][j][k] = -MAXINT;
	//		creat = new list;
	//		creat->x = i + 1;
	//		creat->y = j;
	//		creat->z = k;
	//		creat->next = NULL;
	//		tail->next = creat;
	//		tail = tail->next;
	//	}
	//	if (j > 0 && distance[i][j - 1][k] == MAXINT)
	//	{
	//		distance[i][j - 1][k] = -MAXINT;
	//		creat = new list;
	//		creat->x = i;
	//		creat->y = j - 1;
	//		creat->z = k;
	//		creat->next = NULL;
	//		tail->next = creat;
	//		tail = tail->next;
	//	}
	//	if ((j < gridY - 1) && distance[i][j + 1][k] == MAXINT)
	//	{
	//		distance[i][j + 1][k] = -MAXINT;
	//		creat = new list;
	//		creat->x = i;
	//		creat->y = j + 1;
	//		creat->z = k;
	//		creat->next = NULL;
	//		tail->next = creat;
	//		tail = tail->next;
	//	}
	//	if (k > 0 && distance[i][j][k - 1] == MAXINT)
	//	{
	//		distance[i][j][k - 1] = -MAXINT;
	//		creat = new list;
	//		creat->x = i;
	//		creat->y = j;
	//		creat->z = k - 1;
	//		creat->next = NULL;
	//		tail->next = creat;
	//		tail = tail->next;
	//	}
	//	if ((k < gridZ - 1) && distance[i][j][k + 1] == MAXINT)
	//	{
	//		distance[i][j][k + 1] = -MAXINT;
	//		creat = new list;
	//		creat->x = i;
	//		creat->y = j;
	//		creat->z = k + 1;
	//		creat->next = NULL;
	//		tail->next = creat;
	//		tail = tail->next;
	//	}
	//	//ɾheadĵһԪأΪԪѾˣͷſռ
	//	creat = head;
	//	head = head->next;
	//	delete creat;
	//}


	//for (j = 0 ;j < gridY; j ++)
	//{
	//	for (i = 0; i < gridX; i ++)
	//	{
	//		for (k = 0; k < gridZ; k ++)
	//			fprintf(fout, "%f ", distance[i][j][k]);
	//	fprintf(fout, "\n");
	//	}
	//	fprintf(fout,"\n");
	//}
	//fprintf(fout, "%f ", distance[8][0][8]);

	//distance㣨ۼӣdiscomfortֵ
	for (i = 0; i < gridX; i ++)
		for (j = 0; j < gridY; j ++)
			for (k = 0; k < gridZ; k ++)
			{
				if (fabs(distance[i][j][k]) < (GridLength/* * 1.414*/ / 1.0)) //˵mesh߽
				{
					m_gridData[i][j][k].discomfort += DisValue;
				}
				//else if (distance[i][j][k] != -MAXINT) //˵meshⲿmesh߽
				//{
				//	//ֻҪڵһֵ˵meshռݣDiscomfortӦóɽϴֵ
				//	if (   (i > 0 && distance[i - 1][j][k] > 0) || (i < gridX - 1 && distance[i + 1][j][k] > 0)
				//		|| (j > 0 && distance[i][j - 1][k] > 0) || (j < gridY - 1 && distance[i][j + 1][k] > 0)
				//		|| (k > 0 && distance[i][j][k - 1] > 0) || (k < gridZ - 1 && distance[i][j][k + 1] > 0))
				//		m_gridData[i][j][k].discomfort += DISVALUE;
				//}
			}
	return true;
}

void Potential::Vertex_and_edge_Normals(GLMmodel* model,SparseMatrix<Vec3f>& edge_normals)
{
	GLfloat angle_a,angle_b,angle_c;
	GLfloat edge_A,edge_B,edge_C;
	Vec3f point_a,point_b,point_c,face_normal,p;
	GLuint index_a,index_b,index_c;
	GLuint edges_normal_sizes;
	assert(model);
	assert(model->facetnorms);
	if (model->normals) {
		free(model->normals);
	}
	model->numnormals = model->numvertices;
	model->normals = (GLfloat*)malloc(sizeof(GLfloat)* 3* (model->numnormals + 1));
	for (int j = 0;j < 3 * (model->numnormals + 1);j ++)
	{
		model->normals[j]=0.0;
	}
	edges_normal_sizes=model->numvertices;
	edge_normals.resize(edges_normal_sizes);
	edge_normals.zero();
	for (GLuint i=0;i<model->numtriangles;i++)
	{
		index_a=model->triangles[i].vindices[0];
		index_b=model->triangles[i].vindices[1];
		index_c=model->triangles[i].vindices[2];

		point_a[0]=model->vertices[3*index_a];
		point_a[1]=model->vertices[3*index_a+1];
		point_a[2]=model->vertices[3*index_a+2];

		point_b[0]=model->vertices[3*index_b];
		point_b[1]=model->vertices[3*index_b+1];
		point_b[2]=model->vertices[3*index_b+2];

		point_c[0]=model->vertices[3*index_c];
		point_c[1]=model->vertices[3*index_c+1];
		point_c[2]=model->vertices[3*index_c+2];

		face_normal[0]=model->facetnorms[3*(i+1)];
		face_normal[1]=model->facetnorms[3*(i+1)+1];
		face_normal[2]=model->facetnorms[3*(i+1)+2];

		//compute edge_normals increament
		edge_normals.add_to_element(index_a-1,index_b-1,face_normal);
		edge_normals.add_to_element(index_b-1,index_a-1,face_normal);

		edge_normals.add_to_element(index_b-1,index_c-1,face_normal);
		edge_normals.add_to_element(index_c-1,index_b-1,face_normal);

		edge_normals.add_to_element(index_a-1,index_c-1,face_normal);
		edge_normals.add_to_element(index_c-1,index_a-1,face_normal);

		//compute vertex_normal increament
		edge_A=mag(point_b-point_c);
		edge_B=mag(point_a-point_c);
		edge_C=mag(point_a-point_b);
		angle_a=(sqr(edge_B)+sqr(edge_C)-sqr(edge_A))/2.0/edge_B/edge_C;
		angle_a=acos(angle_a);
		angle_b=(sqr(edge_A)+sqr(edge_C)-sqr(edge_B))/2.0/edge_A/edge_C;
		angle_b=acos(angle_b);
		angle_c=(sqr(edge_A)+sqr(edge_B)-sqr(edge_C))/2.0/edge_A/edge_B;
		angle_c=acos(angle_c);
		model->normals[3*index_a]+=angle_a*face_normal[0];
		model->normals[3*index_a+1]+=angle_a*face_normal[1];
		model->normals[3*index_a+2]+=angle_a*face_normal[2];

		model->normals[3*index_b]+=angle_b*face_normal[0];
		model->normals[3*index_b+1]+=angle_b*face_normal[1];
		model->normals[3*index_b+2]+=angle_b*face_normal[2];

		model->normals[3*index_c]+=angle_c*face_normal[0];
		model->normals[3*index_c+1]+=angle_c*face_normal[1];
		model->normals[3*index_c+2]+=angle_c*face_normal[2];
	}
}

//objģͶƬƬƬ㣬ǰ񶥵
float Potential::ComputeDistance(GLMmodel* model, int trangle_index,int index_a,int index_b,int index_c,Vec3f trangle_point_a,Vec3f trangle_point_b,Vec3f trangle_point_c,Vec3f point)
{
	float dis,t;
	Vec3f facenormal,intersection_point,sign_normal;
	float edge_A,edge_B,edge_C;              //߳
	float dis_of_intersectionpoint_to_edgeA,dis_of_intersectionpoint_to_edgeB,dis_of_intersectionpoint_to_edgeC;    //㵽ߵľ
	Vec3f  Stagnation_A,Stagnation_B,Stagnation_C;       //㵽ÿߵĽŵ
	float dis_to_pointA,dis_to_pointB,dis_to_pointC;    //㵽ÿľ
	float trangle_area,sum_area;
	bool flag_edgeA,flag_edgeB,flag_edgeC;
	float temp_disA,temp_disB,temp_disC;
	float sign_A,sign_B,sign_C,sign_edge_A,sign_edge_B,sign_edge_C;
	facenormal[0]=model->facetnorms[3*(trangle_index+1)];
	facenormal[1]=model->facetnorms[3*(trangle_index+1)+1];
	facenormal[2]=model->facetnorms[3*(trangle_index+1)+2];

	t=dot(facenormal,trangle_point_a-point)/mag2(facenormal);
	intersection_point=facenormal*t+point;

	edge_A=mag(trangle_point_b-trangle_point_c);
	edge_B=mag(trangle_point_a-trangle_point_c);
	edge_C=mag(trangle_point_a-trangle_point_b);

	t=dot(trangle_point_b-trangle_point_c,point-trangle_point_b)/mag2(trangle_point_b-trangle_point_c);
	Stagnation_A=(trangle_point_b-trangle_point_c)*t+trangle_point_b;

	t=dot(trangle_point_a-trangle_point_c,point-trangle_point_a)/mag2(trangle_point_a-trangle_point_c);
	Stagnation_B=(trangle_point_a-trangle_point_c)*t+trangle_point_a;

	t=dot(trangle_point_a-trangle_point_b,point-trangle_point_b)/mag2(trangle_point_a-trangle_point_b);
	Stagnation_C=(trangle_point_a-trangle_point_b)*t+trangle_point_b;

	dis_of_intersectionpoint_to_edgeA=mag(intersection_point-Stagnation_A);
	dis_of_intersectionpoint_to_edgeB=mag(intersection_point-Stagnation_B);
	dis_of_intersectionpoint_to_edgeC=mag(intersection_point-Stagnation_C);

	trangle_area=sqrt((edge_A+edge_B+edge_C)*(edge_A+edge_B-edge_C)*(edge_B+edge_C-edge_A)*(edge_A+edge_C-edge_B))/4.0;
	sum_area=(edge_A*dis_of_intersectionpoint_to_edgeA+edge_B*dis_of_intersectionpoint_to_edgeB+edge_C*dis_of_intersectionpoint_to_edgeC)/2.0;

	dis_to_pointA=mag(point-trangle_point_a);
	dis_to_pointB=mag(point-trangle_point_b);
	dis_to_pointC=mag(point-trangle_point_c);

	//ڶA
	sign_normal[0]=model->normals[3*index_a];
	sign_normal[1]=model->normals[3*index_a+1];
	sign_normal[2]=model->normals[3*index_a+2];
	sign_A=inside_or_outside(point-trangle_point_a,sign_normal);
	//ڶB
	sign_normal[0]=model->normals[3*index_b];
	sign_normal[1]=model->normals[3*index_b+1];
	sign_normal[2]=model->normals[3*index_b+2];
	sign_B=inside_or_outside(point-trangle_point_b,sign_normal);
	//ڶC
	sign_normal[0]=model->normals[3*index_c];
	sign_normal[1]=model->normals[3*index_c+1];
	sign_normal[2]=model->normals[3*index_c+2];
	sign_C=inside_or_outside(point-trangle_point_c,sign_normal);
	//ڱA
	sign_normal=get_SparseMatrix_value(index_b-1,index_c-1);
	sign_edge_A=inside_or_outside(point-Stagnation_A,sign_normal);
	//ڱB
	sign_normal=get_SparseMatrix_value(index_a-1,index_c-1);
	sign_edge_B=inside_or_outside(point-Stagnation_B,sign_normal);
	//ڱC
	sign_normal=get_SparseMatrix_value(index_b-1,index_a-1);
	sign_edge_C=inside_or_outside(point-Stagnation_C,sign_normal);

	if (((trangle_area-sum_area)<=MININT)&&((trangle_area-sum_area)>=-MININT))
	{
		dis=mag(point-intersection_point);
		//㲻ڱߺͶ
		if (dis_of_intersectionpoint_to_edgeA*dis_of_intersectionpoint_to_edgeB*dis_of_intersectionpoint_to_edgeC>MININT)
		{
			dis=dis*inside_or_outside(point-intersection_point,facenormal);
			return dis;
		}
		//ڶC
		if ((dis_of_intersectionpoint_to_edgeA<=MININT)&&
			(dis_of_intersectionpoint_to_edgeA>=-MININT)&&(dis_of_intersectionpoint_to_edgeB<=MININT)&&(dis_of_intersectionpoint_to_edgeB>=-MININT))
		{
			dis=dis*sign_C;
			return dis;
		}
		//ڶA
		if ((dis_of_intersectionpoint_to_edgeB<=MININT)&&
			(dis_of_intersectionpoint_to_edgeB>=-MININT)&&(dis_of_intersectionpoint_to_edgeC<=MININT)&&(dis_of_intersectionpoint_to_edgeC>=-MININT))
		{
			dis=dis*sign_A;
			return dis;
		}
		//ڶB
		if ((dis_of_intersectionpoint_to_edgeA<=MININT)&&
			(dis_of_intersectionpoint_to_edgeA>=-MININT)&&(dis_of_intersectionpoint_to_edgeC<=MININT)&&(dis_of_intersectionpoint_to_edgeC>=-MININT))
		{
			dis=dis*sign_B;
			return dis;
		}
		//ڱA
		if ((dis_of_intersectionpoint_to_edgeA<=MININT)&&(dis_of_intersectionpoint_to_edgeA>=-MININT))
		{
			dis=dis*sign_edge_A;
			return dis;
		}
		//ڱB
		if ((dis_of_intersectionpoint_to_edgeB<=MININT)&&(dis_of_intersectionpoint_to_edgeB>=-MININT))
		{
			dis=dis*sign_edge_B;
			return dis;
		}
		//ڱC
		if ((dis_of_intersectionpoint_to_edgeC<=MININT)&&(dis_of_intersectionpoint_to_edgeC>=-MININT))
		{
			dis=dis*sign_edge_C;
			return dis;
		} 
	}
	else
	{
		flag_edgeA=Is_point_on_segment(Stagnation_A,trangle_point_b,trangle_point_c);
		flag_edgeB=Is_point_on_segment(Stagnation_B,trangle_point_a,trangle_point_c);
		flag_edgeC=Is_point_on_segment(Stagnation_C,trangle_point_a,trangle_point_b);
		temp_disA=temp_disB=temp_disC=1000.0;
		if (flag_edgeA)
		{
			temp_disA=mag(point-Stagnation_A);
		}
		if (flag_edgeB)
		{
			temp_disB=mag(point-Stagnation_B);
		}
		if (flag_edgeC)
		{
			temp_disC=mag(point-Stagnation_C);
		}
		dis=min(temp_disA,temp_disB,temp_disC,dis_to_pointA,dis_to_pointB,dis_to_pointC);
		if ((mag(point-intersection_point)<=MININT)&&(mag(point-intersection_point)>=-MININT))
		{
			return -dis;
		}
		if (dis==dis_to_pointA)
		{
			dis=dis*sign_A;
			return dis;
		}
		if (dis==dis_to_pointB)
		{
			dis=dis*sign_B;
			return dis;
		}
		if (dis==dis_to_pointC)
		{
			dis=dis*sign_C;
			return dis;
		}
		if (dis==temp_disA)
		{
			dis=dis*sign_edge_A;
			return dis;
		}
		if (dis==temp_disB)
		{
			dis=dis*sign_edge_B;
			return dis;
		}
		if(dis==temp_disC)
		{
			dis=dis*sign_edge_C;
			return dis;
		}
	}
}

Vec3f Potential::get_SparseMatrix_value(int pointA_index,int pointB_index)
{
	int index;
	for (int i=0;i<edge_normals.index[pointA_index].size();i++)
	{
		if (edge_normals.index[pointA_index][i]==pointB_index)
		{
			index=i;
		}

	}
	return edge_normals.value[pointA_index][index];
}

float Potential::inside_or_outside(Vec3f point, Vec3f signormal)
{
	if (dot(point,signormal)>0)
	{
		return -1.0;
	}
	else
	{
		return 1.0;
	}
}
bool Potential::Is_point_on_segment(Vec3f point,Vec3f pointA,Vec3f pointB)
{
	if (dot(point-pointA,point-pointB)>0)
	{
		return false;
	}
	else
	{
		return true;
	}
}