//Potential.cpp 
#include "Potential.h"
#include "interp.h"
#include <cmath>

Potential::Potential(int x, int y, int z) : gridX(x), gridY(y), gridZ(z) 
{
}

void Potential::SetGoal(int xa, int xb, int ya, int yb, int za, int zb)
{
	m_GoalXa = xa;
	m_GoalXb = xb;
	m_GoalYa = ya;
	m_GoalYb = yb;
	m_GoalZa = za;
	m_GoalZb = zb;
}

void Potential::SetSpeedRange(float _minSpeed, float _maxSpeed)
{
	maxSpeed = _maxSpeed;
	minSpeed = _minSpeed;
}

void Potential::SetDensityRange(float _minDensity, float _maxDensity)
{
	maxDensity = _maxDensity;
	minDensity = _minDensity;
}

void Potential::SetLamdaWeights(float _lamda, float _w1, float _w2, float _w3)
{
	lamda = _lamda;
	w1 = _w1;
	w2 = _w2;
	w3 = _w3;
}

void Potential::ComputeFaceData(void)
{
	int x, y, z;
	float speedT = minSpeed, speedV;
	for(x = 0; x < gridX; x ++)
		for(y = 0; y < gridY; y ++)
			for (z = 0; z < gridZ; z ++)
			{
				if (x + 1 < gridX)
				{
					speedV =  averagev[x + 1][y][z][0];
					if(speedV > speedT) speedV = speedT;
					if(density[x + 1][y][z] <= minDensity)
					{
						m_gridData[x][y][z].faceData[0].speed = speedT;
					}
					else if(density[x + 1][y][z] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[0].speed = speedV;
					}
					else
					{
						m_gridData[x][y][z].faceData[0].speed = speedT + (density[x + 1][y][z] - minDensity)
							/(maxDensity - minDensity) * (speedV-speedT);
					}
				}
				if (y + 1 < gridY)
				{
					speedV =  abs(averagev[x][y + 1][z][1]);
					if(speedV > speedT) speedV = speedT;
					if(density[x][y + 1][z] <= minDensity)
					{
						m_gridData[x][y][z].faceData[1].speed = speedT;
					}
					else if(density[x][y + 1][z] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[1].speed = speedV;
					}
					else
					{
						m_gridData[x][y][z].faceData[1].speed = speedT + (density[x][y + 1][z] - minDensity)
							/(maxDensity - minDensity) * (speedV-speedT);
					}
				}
				if (z + 1 < gridZ)
				{
					speedV =  abs(averagev[x][y][z + 1][2]);
					if(speedV > speedT) speedV = speedT;
					if(density[x][y][z + 1] <= minDensity)
					{
						m_gridData[x][y][z].faceData[2].speed = speedT;
					}
					else if(density[x][y][z + 1] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[2].speed = speedV;
					}
					else
					{
						m_gridData[x][y][z].faceData[2].speed = speedT + (density[x][y][z + 1] - minDensity)
							/(maxDensity - minDensity) * (speedV-speedT);
					}
				}
				if (x - 1 >= 0)
				{
					speedV =  -averagev[x - 1][y][z][0];
					if(speedV < 0) speedV = 0;
					if(speedV > speedT) speedV = speedT;
					if(density[x - 1][y][z] <= minDensity)
					{
						m_gridData[x][y][z].faceData[3].speed = speedT;
					}
					else if(density[x - 1][y][z] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[3].speed = speedV;
					}
					else
					{ 
						m_gridData[x][y][z].faceData[3].speed = speedT + (density[x - 1][y][z] - minDensity) 
							/ (maxDensity - minDensity) * (speedV - speedT);
					}
				}
				if (y - 1 >= 0)
				{
					speedV =  -averagev[x][y - 1][z][1];
					if(speedV < 0) speedV = 0;
					if(speedV > speedT) speedV = speedT;
					if(density[x][y - 1][z] <= minDensity)
					{
						m_gridData[x][y][z].faceData[4].speed = speedT;
					}
					else if(density[x][y - 1][z] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[4].speed = speedV;
					}
					else
					{ 
						m_gridData[x][y][z].faceData[4].speed = speedT + (density[x][y - 1][z] - minDensity) 
							/ (maxDensity - minDensity) * (speedV - speedT);
					}
				}
				if(z - 1 >= 0)
				{
					speedV =  -averagev[x][y][z - 1][2];
					if(speedV < 0) speedV = 0;
					if(speedV > speedT) speedV = speedT;
					if(density[x][y][z - 1] <= minDensity)
					{
						m_gridData[x][y][z].faceData[5].speed = speedT;
					}
					else if(density[x][y][z - 1] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[5].speed = speedV;
					}
					else
					{ 
						m_gridData[x][y][z].faceData[5].speed = speedT + (density[x][y][z - 1] - minDensity) 
							/ (maxDensity - minDensity) * (speedV - speedT);
					}
				}
			}

	for(x = 0; x < gridX; x ++)
		for (y = 0; y < gridY; y ++) 
			for( z = 0; z < gridZ; z ++)
			{
				if(x + 1 < gridX)
				{	
					if(m_gridData[x][y][z].faceData[0].speed <= 0) 
						m_gridData[x][y][z].faceData[0].speed = MININT;
					m_gridData[x][y][z].faceData[0].cost = (w1 * m_gridData[x][y][z].faceData[0].speed + w2 +
						w3 * m_gridData[x + 1][y][z].discomfort) / m_gridData[x][y][z].faceData[0].speed;
				}
				if(y + 1 < gridY)
				{	
					if(m_gridData[x][y][z].faceData[1].speed <= 0) 
						m_gridData[x][y][z].faceData[1].speed = MININT;
					m_gridData[x][y][z].faceData[1].cost = (w1 * m_gridData[x][y][z].faceData[1].speed + w2 +
						w3 * m_gridData[x][y + 1][z].discomfort) / m_gridData[x][y][z].faceData[1].speed;
				}
				if(z + 1 < gridZ)
				{	
					if(m_gridData[x][y][z].faceData[2].speed <= 0) 
						m_gridData[x][y][z].faceData[2].speed = MININT;
					m_gridData[x][y][z].faceData[2].cost = (w1 * m_gridData[x][y][z].faceData[2].speed + w2 +
						w3 * m_gridData[x][y][z + 1].discomfort) / m_gridData[x][y][z].faceData[2].speed;
				}

				if(x - 1 >= 0)
				{
					if(m_gridData[x][y][z].faceData[3].speed <= 0) 
						m_gridData[x][y][z].faceData[3].speed = MININT;
					m_gridData[x][y][z].faceData[3].cost = (w1 * m_gridData[x][y][z].faceData[3].speed + w2 + 
						w3 * m_gridData[x - 1][y][z].discomfort) / m_gridData[x][y][z].faceData[3].speed;
				}
				if(y - 1 >= 0)
				{
					if(m_gridData[x][y][z].faceData[4].speed <= 0) 
						m_gridData[x][y][z].faceData[4].speed = MININT;
					m_gridData[x][y][z].faceData[4].cost = (w1 * m_gridData[x][y][z].faceData[4].speed + w2 + 
						w3 * m_gridData[x][y - 1][z].discomfort) / m_gridData[x][y][z].faceData[4].speed;
				}
				if(z - 1 >= 0)
				{
					if(m_gridData[x][y][z].faceData[5].speed <= 0) 
						m_gridData[x][y][z].faceData[5].speed = MININT;
					m_gridData[x][y][z].faceData[5].cost = (w1 * m_gridData[x][y][z].faceData[5].speed + w2 + 
						w3 * m_gridData[x][y][z - 1].discomfort) / m_gridData[x][y][z].faceData[5].speed;
				}
			}
}

//processڵtargetΪCANDIDATE״̬CandidateУPotential
void Potential::process(int x, int y, int z, priority_queue<Node, vector<Node>, cmp>& Candidate)
{
	if (m_gridData[x][y][z].FMMFlag != UNKNOWN)
		return;

	CalcPotential(x, y, z);

	m_gridData[x][y][z].FMMFlag = CANDIDATE;
	Node tmp;
	tmp.x = x;
	tmp.y = y;
	tmp.z = z;
	tmp.value = m_gridData[x][y][z].potential;
	Candidate.push(tmp);
}
void Potential::FMM()
{
	int totalGrids = gridX * gridY * gridZ;//
	priority_queue<Node, vector<Node>, cmp> Candidate;
	int x, y, z;
	int KnownIndex = 0;


	//ʼ еPotentialΪ״̬ΪUNKOWN
	for(x = 0; x < gridX; x ++)
		for (y = 0; y < gridY; y ++)
			for(z = 0; z < gridZ; z ++)
			{
				m_gridData[x][y][z].FMMFlag = UNKNOWN;
				m_gridData[x][y][z].potential = MAXINT;
			}

	//ĿPotentialΪ0״̬ΪKNOWNʼCandidateѡ
	for(x = m_GoalXa; x <= m_GoalXb; x ++)
		for (y = m_GoalYa; y <= m_GoalYb; y ++)
			for(z = m_GoalZa; z <= m_GoalZb; z ++)
			{
				m_gridData[x][y][z].FMMFlag = KNOWN;
				m_gridData[x][y][z].potential = 0;
				KnownIndex ++;
				if (x + 1 < gridX) process(x + 1, y, z, Candidate);
				if (y + 1 < gridY) process(x, y + 1, z, Candidate);
				if (z + 1 < gridZ) process(x, y, z + 1, Candidate);

				if (x - 1 >= 0) process(x - 1, y, z, Candidate);
				if (y - 1 >= 0) process(x, y - 1, z, Candidate);
				if (z - 1 >= 0) process(x, y, z - 1, Candidate);
			}

	/*KnownIndexδʱѭȰCandidatePotentialֵСĽڵΪKNOWNýڵھӣ㣩
	  CandidateвPotentialʽ11*/
	while (KnownIndex < totalGrids)
	{
		//ȡ׵ԪأڵֵΪKNOWN
		Node tmp = Candidate.top();
		Candidate.pop();
		x = tmp.x;
		y = tmp.y;
		z = tmp.z;
		//ǰڵСֵȶдŵֵͬ˵
		//ýڵѾpopˣѾõСֵٴ
		//if (tmp.value != m_gridData[x][y][z].potential)
			//continue;

		m_gridData[x][y][z].FMMFlag = KNOWN;
		KnownIndex ++;
		if (x + 1 < gridX) process(x + 1, y, z, Candidate);
		if (y + 1 < gridY) process(x, y + 1, z, Candidate);
		if (z + 1 < gridZ) process(x, y, z + 1, Candidate);

		if (x - 1 >= 0) process(x - 1, y, z, Candidate);
		if (y - 1 >= 0) process(x, y - 1, z, Candidate);
		if (z - 1 >= 0) process(x, y, z - 1, Candidate);
	}

	return;
}
void Potential::solve(float &x, float a, float b, float c, float d)
{
	float A = d + b;
	float B = - 2 * a * d - 2 * c * b;
	float C = a * a * d + c * c * b - b * d;
	float D = B * B - 4 * A * C;
	if(D < 0) D = 0;
	x = (- B + sqrt(D)) / (2.0 * A);
}
void Potential::CalcPotential(int x, int y, int z)
{
	float totalmin = MAXINT;
	bool minx = false;
	float a, b = 0, c, d = 0, e ,f = 0;
	if (x < gridX - 1 && m_gridData[x + 1][y][z].potential + m_gridData[x][y][z].faceData[0].cost < totalmin)
	{
		a = m_gridData[x + 1][y][z].potential;
		b = m_gridData[x][y][z].faceData[0].cost;
		totalmin = a + b;
		minx = true; 
	}
	if (x > 0 && m_gridData[x - 1][y][z].potential + m_gridData[x][y][z].faceData[3].cost < totalmin)
	{
		a = m_gridData[x - 1][y][z].potential;
		b = m_gridData[x][y][z].faceData[3].cost;
		minx = true;
	}

	totalmin = MAXINT;
	bool miny = false;
	if (y < gridY - 1 && m_gridData[x][y + 1][z].potential + m_gridData[x][y][z].faceData[1].cost < totalmin)
	{
		c = m_gridData[x][y + 1][z].potential;
		d = m_gridData[x][y][z].faceData[1].cost;
		totalmin = c + d;
		miny = true;
	}
	if (y > 0 && m_gridData[x][y - 1][z].potential + m_gridData[x][y][z].faceData[4].cost < totalmin)
	{
		c = m_gridData[x][y - 1][z].potential;
		d = m_gridData[x][y][z].faceData[4].cost;
		miny = true;
	}

	totalmin = MAXINT;
	bool minz = false;
	if (z < gridZ - 1 && m_gridData[x][y][z + 1].potential + m_gridData[x][y][z].faceData[2].cost < totalmin)
	{
		e = m_gridData[x][y][z + 1].potential;
		f = m_gridData[x][y][z].faceData[2].cost;
		totalmin = e + f;
		minz = true;
	}
	if (z > 0 && m_gridData[x][y][z - 1].potential + m_gridData[x][y][z].faceData[5].cost < totalmin)
	{
		e = m_gridData[x][y][z - 1].potential;
		f = m_gridData[x][y][z].faceData[5].cost;
		minz = true;
	}

	//㹫ʽ11.㷨Ҫ㲻֪дôuglyĳȫelse if
	if (!minx && !miny)
		m_gridData[x][y][z].potential = e + f;
	else if (!minx && !minz)
		m_gridData[x][y][z].potential = c + d;
	else if (!miny && !minz)
		m_gridData[x][y][z].potential = a + b;
	else 
	{
		//ƽü㡣
		b = b * b;
		d = d * d;
		f = f * f;

		if (!minx)
			solve(m_gridData[x][y][z].potential, c, d, e, f);
		else if (!miny)
			solve(m_gridData[x][y][z].potential, a, b, e, f);
		else if (!minz)
			solve(m_gridData[x][y][z].potential, a, b, c, d);
		else
		{
			//ߵһ˵·̫ˡ
			float A = d * f + b * f + b * d;
			float B = - 2 * d * f * a - 2 * b * f * c - 2 * b * d * e;
			float C = a * a * d * f + c * c * b * f + e * e * b * d - b * d * f;
			float D = B * B - 4 * A * C;
			if (D < 0) D = 0;
			m_gridData[x][y][z].potential = (- B + sqrt(D)) / (2.0 * A);
		}

		//if (!minx && !miny && !minz)
		//{
		//	minx = 0;
		//}
	}
}
void Potential::addDensity(Vec3f position, Vec3f velocity)
{
	float x = position[0] / GRIDL;
	float y = position[1] / GRIDL;
	float z = position[2] / GRIDL;
	int i = floor(x + 0.5);
	int j = floor(y + 0.5);
	int k = floor(z + 0.5);
	//Ӧsection 4.1㷨
	if (i > gridX || j > gridY || k > gridZ || i < 0 || j < 0 || k < 0) return;

	float deltax = x - i + 0.5 * GRIDL;
	float deltay = y - j + 0.5 * GRIDL;
	float deltaz = z - k + 0.5 * GRIDL;
	float tmp;
	if (i > 0 && j > 0 && k < gridZ)  //density A
	{
		tmp = min(1 - deltax, 1 - deltay, deltaz);
		tmp = pow(tmp, lamda);
		density[i - 1][j - 1][k] += tmp;
		averagev[i - 1][j - 1][k] += velocity * tmp;
	}
	if (i < gridX && j > 0 && k < gridZ) //density B
	{
		tmp = min(deltax, 1 - deltay, deltaz);
		tmp = pow(tmp, lamda);
		density[i][j - 1][k] += tmp;
		averagev[i][j - 1][k] += velocity * tmp;
	}
	if (i < gridX && j < gridY && k < gridZ) //density C
	{
		tmp = min(deltax, deltay, deltaz);
		tmp = pow(tmp, lamda);
		density[i][j][k] += tmp;
		averagev[i][j][k] += velocity * tmp;
	}
	if (i > 0 && j < gridY && k < gridZ) //density D
	{
		tmp = min(1 - deltax, deltay, deltaz);
		tmp = pow(tmp, lamda);
		density[i - 1][j][k] += tmp;
		averagev[i - 1][j][k] += velocity * tmp;
	}

	if (i > 0 && j > 0 && k > 0)  //density E
	{
		tmp = min(1 - deltax, 1 - deltay, 1 - deltaz);
		tmp = pow(tmp, lamda);
		density[i - 1][j - 1][k - 1] += tmp;
		averagev[i - 1][j - 1][k - 1] += velocity * tmp;
	}
	if (i < gridX && j > 0 && k > 0) //density F
	{
		tmp = min(deltax, 1 - deltay, 1 - deltaz);
		tmp = pow(tmp, lamda);
		density[i][j - 1][k - 1] += tmp;
		averagev[i][j - 1][k - 1] += velocity * tmp;
	}
	if (i < gridX && j < gridY && k > 0) //density G
	{
		tmp = min(deltax, deltay, 1 - deltaz);
		tmp = pow(tmp, lamda);
		density[i][j][k - 1] += tmp;
		averagev[i][j][k - 1] += velocity * tmp;
	}
	if (i > 0 && j < gridY && k > 0) //density H
	{
		tmp = min(1 - deltax, deltay, 1 - deltaz);
		tmp = pow(tmp, lamda);
		density[i - 1][j][k - 1] += tmp;
		averagev[i - 1][j][k - 1] += velocity * tmp;
	}
}
Vec3f Potential::GetVelocity(Vec3f position)
{
	//ֵgradient
	Vec3f gradient = get_PotentialGradient(position);
	gradient = normalized(-gradient);

	//i,j,kʾposition
	int i = floor(position[0] / GRIDL);
	int j = floor(position[1] / GRIDL);
	int k = floor(position[2] / GRIDL);
	if (m_gridData[i][j][k].potential == 0) 
		return(Vec3f(0, 0, 0));
	//x,y,zʾpositiongradientλ
	int x = floor(position[0] / GRIDL + gradient[0] / 2 + 0.5);
	int y = floor(position[1] / GRIDL + gradient[1] / 2 + 0.5);
	int z = floor(position[2] / GRIDL + gradient[2] / 2 + 0.5);
	float speed = m_gridData[i][j][k].faceData[(-3) * (x - i) + 3].speed * abs(gradient[0]) + 
		m_gridData[i][j][k].faceData[(-3) * (y - j) + 4].speed * abs(gradient[1]) +
		m_gridData[i][j][k].faceData[(-3) * (z - k) + 5].speed * abs(gradient[2]);
	Vec3f velocity = gradient * speed;
	return velocity;
}
Vec3f Potential::get_PotentialGradient(Vec3f position)
{
	int i,j,k;
	float fx,fy,fz;
	Vec3f gradient;

	get_barycentric(position[0], i, fx, 0, gridX);
	get_barycentric(position[1], j, fy, 0, gridY);
	get_barycentric(position[2], k, fz, 0, gridZ);
	//dx
	float v000 = (m_gridData[i+1][j][k].potential - m_gridData[i-1][j][k].potential) / GRIDL * 0.5f;
	float v001 = (m_gridData[i+1][j][k+1].potential - m_gridData[i-1][j][k+1].potential) / GRIDL * 0.5f;
	float v010 = (m_gridData[i+1][j+1][k].potential - m_gridData[i-1][j+1][k].potential) / GRIDL*0.5f;
	float v011 = (m_gridData[i+1][j+1][k+1].potential - m_gridData[i-1][j+1][k+1].potential) / GRIDL * 0.5f;
	float v100 = (m_gridData[i+2][j][k].potential - m_gridData[i][j][k].potential) / GRIDL * 0.5f;
	float v101 = (m_gridData[i+2][j][k+1].potential-m_gridData[i][j][k+1].potential) / GRIDL * 0.5f;
	float v110 = (m_gridData[i+2][j+1][k].potential-m_gridData[i][j+1][k].potential) / GRIDL * 0.5f;
	float v111 = (m_gridData[i+2][j+1][k+1].potential - m_gridData[i][j+1][k+1].potential) / GRIDL * 0.5f;
	gradient[0]=trilerp(v000, v100, v010, v110, v001, v101, v011, v111, fx, fy, fz);
	//dy
	v000 = (m_gridData[i][j+1][k].potential-m_gridData[i][j-1][k].potential)/GRIDL*0.5f;
	v001 = (m_gridData[i][j+1][k+1].potential-m_gridData[i][j-1][k+1].potential)/GRIDL*0.5f;
	v010 = (m_gridData[i][j+2][k].potential-m_gridData[i][j][k].potential)/GRIDL*0.5f;
	v011 = (m_gridData[i][j+2][k+1].potential-m_gridData[i][j][k+1].potential)/GRIDL*0.5f;
	v100 = (m_gridData[i+1][j+1][k].potential-m_gridData[i+1][j-1][k].potential)/GRIDL*0.5f;
	v101 = (m_gridData[i+1][j+1][k+1].potential-m_gridData[i+1][j-1][k+1].potential)/GRIDL*0.5f;
	v110 = (m_gridData[i+1][j+2][k].potential-m_gridData[i+1][j][k].potential)/GRIDL*0.5f;
	v111 = (m_gridData[i+1][j+2][k+1].potential-m_gridData[i+1][j][k+1].potential)/GRIDL*0.5f;
	gradient[1]=trilerp(v000, v100,v010, v110, v001, v101,v011, v111,fx, fy, fz);
	//dz
	v000 = (m_gridData[i][j][k+1].potential-m_gridData[i][j][k-1].potential)/GRIDL*0.5f;
	v001 = (m_gridData[i][j][k+2].potential-m_gridData[i][j][k].potential)/GRIDL*0.5f;
	v010 = (m_gridData[i][j+1][k+1].potential-m_gridData[i][j+1][k-1].potential)/GRIDL*0.5f;
	v011 = (m_gridData[i][j+1][k+2].potential-m_gridData[i][j+1][k].potential)/GRIDL*0.5f;
	v100 = (m_gridData[i+1][j][k+1].potential-m_gridData[i+1][j][k-1].potential)/GRIDL*0.5f;
	v101 = (m_gridData[i+1][j][k+2].potential-m_gridData[i+1][j][k].potential)/GRIDL*0.5f;
	v110 = (m_gridData[i+1][j+1][k+1].potential-m_gridData[i+1][j+1][k-1].potential)/GRIDL*0.5f;
	v111 = (m_gridData[i+1][j+1][k+2].potential-m_gridData[i+1][j+1][k].potential)/GRIDL*0.5f;
	gradient[2]=trilerp(v000, v100,v010, v110, v001, v101,v011, v111,fx, fy, fz);

	return gradient;
}

Vec3f Potential::get_DensityGradient(Vec3f position)
{
	int i,j,k;
	float fx,fy,fz;
	Vec3f gradient;

	get_barycentric(position[0], i, fx, 0, gridX);
	get_barycentric(position[1], j, fy, 0, gridY);
	get_barycentric(position[2], k, fz, 0, gridZ);
	//dx
	float v000 = (density[i+1][j][k] - density[i-1][j][k])* 0.5f;
	float v001 = (density[i+1][j][k+1] - density[i-1][j][k+1])* 0.5f;
	float v010 = (density[i+1][j+1][k] - density[i-1][j+1][k])*0.5f;
	float v011 = (density[i+1][j+1][k+1] - density[i-1][j+1][k+1])* 0.5f;
	float v100 = (density[i+2][j][k] - density[i][j][k])* 0.5f;
	float v101 = (density[i+2][j][k+1]-density[i][j][k+1])* 0.5f;
	float v110 = (density[i+2][j+1][k]-density[i][j+1][k])* 0.5f;
	float v111 = (density[i+2][j+1][k+1] - density[i][j+1][k+1])* 0.5f;
	gradient[0]=trilerp(v000, v100, v010, v110, v001, v101, v011, v111, fx, fy, fz);
	//dy
	v000 = (density[i][j+1][k]-density[i][j-1][k])*0.5f;
	v001 = (density[i][j+1][k+1]-density[i][j-1][k+1])*0.5f;
	v010 = (density[i][j+2][k]-density[i][j][k])*0.5f;
	v011 = (density[i][j+2][k+1]-density[i][j][k+1])*0.5f;
	v100 = (density[i+1][j+1][k]-density[i+1][j-1][k])*0.5f;
	v101 = (density[i+1][j+1][k+1]-density[i+1][j-1][k+1])*0.5f;
	v110 = (density[i+1][j+2][k]-density[i+1][j][k])*0.5f;
	v111 = (density[i+1][j+2][k+1]-density[i+1][j][k+1])*0.5f;
	gradient[1]=trilerp(v000, v100,v010, v110, v001, v101,v011, v111,fx, fy, fz);
	//dz
	v000 = (density[i][j][k+1]-density[i][j][k-1])*0.5f;
	v001 = (density[i][j][k+2]-density[i][j][k])*0.5f;
	v010 = (density[i][j+1][k+1]-density[i][j+1][k-1])*0.5f;
	v011 = (density[i][j+1][k+2]-density[i][j+1][k])*0.5f;
	v100 = (density[i+1][j][k+1]-density[i+1][j][k-1])*0.5f;
	v101 = (density[i+1][j][k+2]-density[i+1][j][k])*0.5f;
	v110 = (density[i+1][j+1][k+1]-density[i+1][j+1][k-1])*0.5f;
	v111 = (density[i+1][j+1][k+2]-density[i+1][j+1][k])*0.5f;
	gradient[2]=trilerp(v000, v100,v010, v110, v001, v101,v011, v111,fx, fy, fz);

	return gradient;
}

//void Potential::addPotential()
//{
//	if(sum_num<MAXSUMNUM) sum_num++;
//	for(int m=0;m<MAXgridX/*gridX*/;m++)
//		for(int n=0;n<MAXgridZ/*gridZ*/;n++)
//			m_gridData[m][n].p[sum_index] = m_gridData[m][n].potential;
//	sum_index++;
//	if(sum_index>=MAXSUMNUM) sum_index=0;
//	for(int m=0;m<MAXgridX/*gridX*/;m++)
//		for(int n=0;n<MAXgridZ/*gridZ*/;n++)
//		{
//			m_gridData[m][n].ave_potential = 0;
//			for(int i=0;i<sum_num;i++)
//				m_gridData[m][n].ave_potential += m_gridData[m][n].p[i];
//			m_gridData[m][n].ave_potential /= sum_num;
//		}
//}