#include "app.h"
#include "Horde3D.h"
#include "Horde3DUtils.h"
#include <math.h>
#include <iomanip>
#include <GL\glut.h>
#include <cstdio>
#include <iostream>

using namespace std;

// Ƕת
inline float degToRad( float f ) 
{
	return f * (3.1415926f / 180.0f);
}

Potential Particle::m_potential = Potential(20, 20, 20);

// ʼ
Application::Application( const std::string &appPath )
{
	for( unsigned int i = 0; i < 320; ++i )
	{	
		_keys[i] = false;
		_prevKeys[i] = false;
	}

	_x = 15; _y = 20; _z = 15; _rx = -90; _ry = 0; _rz = 0; _velocity = 5.0f;
	_curFPS = 30;
	display_timestep = 1 / _curFPS;
	simulation_timestep = 0.025;

	_statMode = 0;
	_freezeMode = 0; _debugViewMode = false; _wireframeMode = false;
	_cam = 0;

	_contentDir = appPath + "../Content";

}


bool Application::init()
{	
	// Initialize engine
	if( !h3dInit() )
	{	
		h3dutDumpMessages();
		return false;
	}

	//init_day();
	init_lamp();

	//================================
	// Init random generator
	//srand( (unsigned int)time( NULL ) );
	srand( 99777 );  // Use fixed value for better performance comparisons

	//ģͺͶ󶨵horde
	H3DRes characterRes = h3dAddResource( H3DResTypes::SceneGraph, "models/fly/fly.scene.xml", 0 );
	H3DRes characterWalkRes = h3dAddResource( H3DResTypes::Animation, "animations/fly.anim", 0 );
	h3dutLoadResourcesFromDisk( _contentDir.c_str() );

	freopen("swarm.out","w",stdout);

	//ҪĶĺ
	makenoise = new MakeNoise();
	//ô˺ӵ״̬ʼ
	makenoise->seed_particles(particles, 60);


	//趨ʼĿĵأٶȷΧ¶ȷΧ
	gridX = Particle::m_potential.gridX;
	gridY = Particle::m_potential.gridY;
	gridZ = Particle::m_potential.gridZ;

	Particle::m_potential.SetGoal(5, 7, 5, 7, 5, 7);
	Particle::m_potential.SetSpeedRange(0.8,5.0);
	Particle::m_potential.SetDensityRange(60.0, 200.0);
	Particle::m_potential.SetLamdaWeights(1, 0.1, 0.6, 0.2);

	for (iterator i = particles.begin(); i != particles.end(); i ++)
	{
		(*i).getVelocity();
		// Add character to scene and apply animation
		//ӵhorde3dĶ
		(*i).node = h3dAddNodes( H3DRootNode, characterRes );
		//Ӹ趯
		//ʼһֵʾ
		h3dSetupModelAnimStage( (*i).node, 0, characterWalkRes, 0, "", false );
		(*i).animTime = rand() % 60;

		//ӵĵǰλ
		//Ҫת
		h3dSetNodeTransform( (*i).node, (*i).Position[0], (*i).Position[1], (*i).Position[2], 0, 0, 0, 0.3, 0.3, 0.3);
	}
	//ܳ
	Particle::m_potential.ComputeFaceData();
	Particle::m_potential.FMM();
	return true;
}

//updateÿһ֡λá
void Application::update(float fps)
{
	if( ! _freezeMode )
	{
//---------------------begin---------continuum crowds㷨----------------------------------

		//ϵֵpotientaldiscomfortֵ
		resetPotential();

		//ۼƼTotalDataеܶȺٶ section4.1е㷨
		for (iterator i = particles.begin(); i != particles.end(); i++)
			(*i).updateDensity();

		for(int x = 0; x < gridX; x ++)
			for (int y = 0; y < gridY; y ++)
				for(int z = 0; z < gridZ; z++)
				if(Particle::m_potential.density[x][y][z] > 0) 
					Particle::m_potential.averagev[x][y][z] = Particle::m_potential.averagev[x][y][z] / Particle::m_potential.density[x][y][z];

		//ܳ
		//Particle::m_potential.ComputeFaceData();
		//Particle::m_potential.FMM();

//---------------------end-----------continuum crowds㷨----------------------------------

		Vec3f midx;
		//for( unsigned int j = tracelength - 1; j > 0; -- j )
		//	_particles[j] = _particles[j-1];

		display_timestep = 1 / fps;
		int i;
		//ӵٶֵӵĵǰλãnoise_velocity
		for (i = 0; i != particles.size(); i ++)
		{
			particles[i].forward = Vec3f(particles[i].Position);
			// almost RK2 (not updating t halfway through time step, so it's technically only 1st order accurate)
			makenoise -> get_velocity( particles[i].Position, particles[i].noise_velocity );
			particles[i].getGlobalVelocity(particles[i].Position);
			midx =particles[i].Position + 0.5f * display_timestep * particles[i].getVelocity();

			makenoise -> get_velocity( midx, particles[i].noise_velocity );
			particles[i].getGlobalVelocity(midx);
			
			particles[i].Position += display_timestep * particles[i].getVelocity();
			particles[i].forward = Vec3f(particles[i].Position - particles[i].forward);
			normalize(particles[i].forward);
		}

		//ײ Чʺܵͣ
		for (i = 0; i != particles.size(); i ++)
			for (int j = i + 1; j != particles.size(); j ++)
			{
				float d = dist(particles[i].Position,particles[j].Position);
				if (d < (particles[i].radius+particles[j].radius))
				{
					Vec3f fij = particles[i].Position - particles[j].Position;
					Vec3f adjust = normalized(fij) * (particles[i].radius + particles[j].radius);
					adjust = (adjust - fij) / 2.0;
					particles[i].Position -= adjust;
					particles[j].Position += adjust;						
				}
			}

		for (i = 0; i != particles.size(); i ++)
		{

			h3dSetNodeTransform( particles[i].node, particles[i].Position[0], particles[i].Position[1], particles[i].Position[2], 0, acos(particles[i].forward[2]) * 57.29578, asin(particles[i].forward[1]) * 57.29578, 0.3, 0.3, 0.3 );
			particles[i].animTime += 1.8f;  
			h3dSetModelAnimParams( particles[i].node, 0, particles[i].animTime, 1.0f );
		}
		i = 0;
		printf("%.3f,%.3f,%.3f ----------",particles[i].global_velocity[0],particles[i].global_velocity[1],particles[i].global_velocity[2]);
		printf(" %.3f,%.3f,%.3f \n",particles[i].Position[0],particles[i].Position[1],particles[i].Position[2]);

		//psi
		makenoise -> advance_time( simulation_timestep );
	}
}

void Application::resetPotential(void)
{
	for(int x = 0; x < gridX; x ++)
		for (int y = 0; y < gridY; y ++)
			for(int z = 0; z < gridZ; z ++)
			{
				Particle::m_potential.density[x][y][z] = 0;
				Particle::m_potential.averagev[x][y][z] = Vec3f(0,0,0);
				Particle::m_potential.m_gridData[x][y][z].discomfort = 0;
			}
			
}


void Application::init_day()
{
	// Set options
	h3dSetOption( H3DOptions::LoadTextures, 1 );
	h3dSetOption( H3DOptions::TexCompression, 0 );
	h3dSetOption( H3DOptions::MaxAnisotropy, 4 );
	h3dSetOption( H3DOptions::ShadowMapSize, 2048 );
	h3dSetOption( H3DOptions::FastAnimation, 1 );

	// Add resources Щòʡƹ⡢ʾģʽȣԲø
	// Pipelines
	_forwardPipeRes = h3dAddResource( H3DResTypes::Pipeline, "pipelines/forward.pipeline.xml", 0 );
	_deferredPipeRes = h3dAddResource( H3DResTypes::Pipeline, "pipelines/deferred.pipeline.xml", 0 );
	// Overlays
	_fontMatRes = h3dAddResource( H3DResTypes::Material, "overlays/font.material.xml", 0 );
	_panelMatRes = h3dAddResource( H3DResTypes::Material, "overlays/panel.material.xml", 0 );
	// Shader for deferred shading
	H3DRes lightMatRes = h3dAddResource( H3DResTypes::Material, "materials/light.material.xml", 0 );
	// Environment
	H3DRes envRes = h3dAddResource( H3DResTypes::SceneGraph, "models/platform/platform.scene.xml", 0 );
	// Skybox
	H3DRes skyBoxRes = h3dAddResource( H3DResTypes::SceneGraph, "models/highway-sky01/highway-sky01.scene.xml", 0 );

	// Load resources
	h3dutLoadResourcesFromDisk( _contentDir.c_str() );

	// Add scene nodes
	// Add camera
	_cam = h3dAddCameraNode( H3DRootNode, "Camera", _forwardPipeRes );
	h3dSetNodeParamI( _cam, H3DCamera::OccCullingI, 1 );

	// Add environment
	H3DNode env = h3dAddNodes( H3DRootNode, envRes );
	h3dSetNodeTransform( env, 0, 0, 0, 0, 0, 0, 0.8f, 0.8f, 0.8f );
	// Add skybox
	H3DNode sky = h3dAddNodes( H3DRootNode, skyBoxRes );
	h3dSetNodeTransform( sky, 0, 0, 0, 0, 0, 0, 0.03, 0.03, 0.03 );
	h3dSetNodeFlags( sky, H3DNodeFlags::NoCastShadow, true );

	// Add light source
	//ģ䣬棬ϿգǶ180ȣûӰ
	H3DNode light = h3dAddLightNode( H3DRootNode, "Light1", lightMatRes, "LIGHTING", "SHADOWMAP" );
	h3dSetNodeTransform(light, 0, 80, 0, -90, 0, 0, 1, 1, 1 );  //߶80ģ͵ĴС
	h3dSetNodeParamF( light, H3DLight::RadiusF, 0, 2100 );
	h3dSetNodeParamF( light, H3DLight::FovF, 0, 180 );
	h3dSetNodeParamI( light, H3DLight::ShadowMapCountI, 0 );
	h3dSetNodeParamF( light, H3DLight::ShadowSplitLambdaF, 0, 0.5f );
	h3dSetNodeParamF( light, H3DLight::ShadowMapBiasF, 0, 0.001f );
	h3dSetNodeParamF( light, H3DLight::ColorF3, 0, 0.5f );
	h3dSetNodeParamF( light, H3DLight::ColorF3, 1, 0.5f );
	h3dSetNodeParamF( light, H3DLight::ColorF3, 2, 0.5f );

	//add light 2------------------------ 
	//ģֱ̫䣬Ƕ-70
	H3DNode light1 = h3dAddLightNode( H3DRootNode, "Light2", lightMatRes, "LIGHTING", "SHADOWMAP" );
	h3dSetNodeTransform(light1, 0, 60, 40, -70, 0, 0, 1, 1, 1 );//߶80
	h3dSetNodeParamF( light1, H3DLight::RadiusF, 0, 2100 );
	h3dSetNodeParamF( light1, H3DLight::FovF, 0, 90 );
	h3dSetNodeParamI( light1, H3DLight::ShadowMapCountI, 3);
	h3dSetNodeParamF( light1, H3DLight::ShadowSplitLambdaF, 0, 0.9f );
	h3dSetNodeParamF( light1, H3DLight::ShadowMapBiasF, 0, 0.001f );
	h3dSetNodeParamF( light1, H3DLight::ColorF3, 0, 0.6f );
	h3dSetNodeParamF( light1, H3DLight::ColorF3, 1, 0.6f );
	h3dSetNodeParamF( light1, H3DLight::ColorF3, 2, 0.6f );

	//add light 3
	//ģ䣬
	H3DNode light2 = h3dAddLightNode( H3DRootNode, "Light3", lightMatRes, "LIGHTING", "SHADOWMAP" );
	h3dSetNodeTransform(light2, 0, 0, 0, 90, 0, 0, 1, 1, 1 );//(light1, 485, 50, 852, -90, 0, 80, 1, 1, 1 );
	h3dSetNodeParamF( light2, H3DLight::RadiusF, 0, 1000 );
	h3dSetNodeParamF( light2, H3DLight::FovF, 0, 250 );
	h3dSetNodeParamI( light2, H3DLight::ShadowMapCountI, 0);
	h3dSetNodeParamF( light2, H3DLight::ShadowSplitLambdaF, 0, 0.9f );
	h3dSetNodeParamF( light2, H3DLight::ShadowMapBiasF, 0, 0.001f );
	h3dSetNodeParamF( light2, H3DLight::ColorF3, 0, 0.9f );
	h3dSetNodeParamF( light2, H3DLight::ColorF3, 1, 0.9f );
	h3dSetNodeParamF( light2, H3DLight::ColorF3, 2, 0.9f );
}

void Application::init_lamp()
{	
	// Set options
	h3dSetOption( H3DOptions::LoadTextures, 1 );
	h3dSetOption( H3DOptions::TexCompression, 0 );
	h3dSetOption( H3DOptions::MaxAnisotropy, 4 );
	h3dSetOption( H3DOptions::ShadowMapSize, 2048 );
	h3dSetOption( H3DOptions::FastAnimation, 1 );

	// Add resources Щòʡƹ⡢ʾģʽȣԲø
	// Pipelines
	_forwardPipeRes = h3dAddResource( H3DResTypes::Pipeline, "pipelines/forward.pipeline.xml", 0 );
	_deferredPipeRes = h3dAddResource( H3DResTypes::Pipeline, "pipelines/deferred.pipeline.xml", 0 );
	// Overlays
	_fontMatRes = h3dAddResource( H3DResTypes::Material, "overlays/font.material.xml", 0 );
	_panelMatRes = h3dAddResource( H3DResTypes::Material, "overlays/panel.material.xml", 0 );
	// Shader for deferred shading
	H3DRes lightMatRes = h3dAddResource( H3DResTypes::Material, "materials/light.material.xml", 0 );
	// Environment
	H3DRes envRes = h3dAddResource( H3DResTypes::SceneGraph, "models/platform/platform.scene.xml", 0 );
	H3DRes lampRes = h3dAddResource( H3DResTypes::SceneGraph, "models/lamp/lamp.scene.xml", 0 );
	// Skybox
	H3DRes skyBoxRes = h3dAddResource( H3DResTypes::SceneGraph, "models/sky02/sky02.scene.xml", 0 );
	// Load resources
	h3dutLoadResourcesFromDisk( _contentDir.c_str() );

	// Add scene nodes
	// Add camera
	_cam = h3dAddCameraNode( H3DRootNode, "Camera", _forwardPipeRes );
	//h3dSetNodeParamI( _cam, H3DCamera::OccCullingI, 1 );

	// Add environment
	H3DNode env = h3dAddNodes( H3DRootNode, envRes );
	h3dSetNodeTransform( env, 0, 0, 0, 0, 0, 0, 0.8f, 0.8f, 0.8f );
	H3DNode lamp = h3dAddNodes( H3DRootNode, lampRes );
	h3dSetNodeTransform( lamp, 0, 0, 0, 0, 5, 0, 0.3, 0.3, 0.3 );

	// Add skybox
	H3DNode sky = h3dAddNodes( H3DRootNode, skyBoxRes );
	h3dSetNodeTransform( sky, 0, 0, 0, 0, 0, 0, 0.03, 0.03, 0.03 );
	h3dSetNodeFlags( sky, H3DNodeFlags::NoCastShadow, true );

	// Add light source
	//ģ䣬棬ϿգǶ180ȣûӰ
	H3DNode light = h3dAddLightNode( H3DRootNode, "Light1", lightMatRes, "LIGHTING", "SHADOWMAP" );
	h3dSetNodeTransform(light, 0, 80, 0, -90, 0, 0, 1, 1, 1 );  //߶80ģ͵ĴС
	h3dSetNodeParamF( light, H3DLight::RadiusF, 0, 2100 );
	h3dSetNodeParamF( light, H3DLight::FovF, 0, 180 );
	h3dSetNodeParamI( light, H3DLight::ShadowMapCountI, 0 );
	h3dSetNodeParamF( light, H3DLight::ShadowSplitLambdaF, 0, 0.5f );
	h3dSetNodeParamF( light, H3DLight::ShadowMapBiasF, 0, 0.001f );
	h3dSetNodeParamF( light, H3DLight::ColorF3, 0, 0.5f );
	h3dSetNodeParamF( light, H3DLight::ColorF3, 1, 0.5f );
	h3dSetNodeParamF( light, H3DLight::ColorF3, 2, 0.5f );

}


//ȾԲø
void Application::mainLoop( float fps )
{
	_curFPS = fps;

	h3dSetOption( H3DOptions::DebugViewMode, _debugViewMode ? 1.0f : 0.0f );
	h3dSetOption( H3DOptions::WireframeMode, _wireframeMode ? 1.0f : 0.0f );

	if( !_freezeMode )
	{
		update( _curFPS );
	}

	// Set camera parameters
	h3dSetNodeTransform( _cam, _x, _y, _z, _rx ,_ry, _rz, 1, 1, 1 );

	// Show stats
	h3dutShowFrameStats( _fontMatRes, _panelMatRes, _statMode );
	if( _statMode > 0 )
	{
		if( h3dGetNodeParamI( _cam, H3DCamera::PipeResI ) == _forwardPipeRes )
			h3dutShowText( "Pipeline: forward", 0.03f, 0.24f, 0.026f, 1, 1, 1, _fontMatRes );
		else
			h3dutShowText( "Pipeline: deferred", 0.03f, 0.24f, 0.026f, 1, 1, 1, _fontMatRes );
	}

	// Render scene
	h3dRender( _cam );

	// Finish rendering of frame
	h3dFinalizeFrame();

	// Remove all overlays
	h3dClearOverlays();

	// Write all messages to log file
	h3dutDumpMessages();
}

void Application::release()
{
	delete makenoise; 
	makenoise = 0x0; 

	//for (iterator i = particles.begin(); i != particles.end(); i ++)
	//	delete(i);
	particles.clear();

	// Release engine
	h3dRelease();
}


void Application::resize( int width, int height )
{
	// Resize viewport
	h3dSetNodeParamI( _cam, H3DCamera::ViewportXI, 0 );
	h3dSetNodeParamI( _cam, H3DCamera::ViewportYI, 0 );
	h3dSetNodeParamI( _cam, H3DCamera::ViewportWidthI, width );
	h3dSetNodeParamI( _cam, H3DCamera::ViewportHeightI, height );
	
	// Set virtual camera parameters
	h3dSetupCameraView( _cam, 45.0f, (float)width / height, 0.1f, 1000.0f );
	h3dResizePipelineBuffers( _deferredPipeRes, width, height );
	h3dResizePipelineBuffers( _forwardPipeRes, width, height );
}


void Application::keyStateHandler()
{
	// ----------------
	// Key-press events
	// ----------------
	if( _keys[32] && !_prevKeys[32] )  // Space
	{
		if( ++_freezeMode == 3 ) _freezeMode = 0;
	}
	
	if( _keys[260] && !_prevKeys[260] )  // F3
	{
		if( h3dGetNodeParamI( _cam, H3DCamera::PipeResI ) == _forwardPipeRes )
			h3dSetNodeParamI( _cam, H3DCamera::PipeResI, _deferredPipeRes );
		else
			h3dSetNodeParamI( _cam, H3DCamera::PipeResI, _forwardPipeRes );
	}
	
	if( _keys[264] && !_prevKeys[264] )  // F7
		_debugViewMode = !_debugViewMode;

	if( _keys[265] && !_prevKeys[265] )  // F8
		_wireframeMode = !_wireframeMode;
	
	if( _keys[263] && !_prevKeys[263] )  // F6
	{
		_statMode += 1;
		if( _statMode > H3DUTMaxStatMode ) _statMode = 0;
	}

	// --------------
	// Keya----------
	if( _freezeMode != 2 )
	{
		float curVel = _velocity / _curFPS;
		
		if( _keys[287] ) curVel *= 5;	// LShift
		
		if( _keys['W'] )
		{
			// Move forward
			_x -= sinf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
			_y -= sinf( -degToRad( _rx ) ) * curVel;
			_z -= cosf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
		}
		if( _keys['S'] )
		{
			// Move backward
			_x += sinf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
			_y += sinf( -degToRad( _rx ) ) * curVel;
			_z += cosf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
		}
		if( _keys['A'] )
		{
			// Strafe left
			_x += sinf( degToRad( _ry - 90) ) * curVel;
			_z += cosf( degToRad( _ry - 90 ) ) * curVel;
		}
		if( _keys['D'] )
		{
			// Strafe right
			_x += sinf( degToRad( _ry + 90 ) ) * curVel;
			_z += cosf( degToRad( _ry + 90 ) ) * curVel;
		}
	}
}


void Application::mouseMoveEvent( float dX, float dY )
{
	if( _freezeMode == 2 ) return;
	
	// Look left/right
	_ry -= dX / 100 * 30;
	
	// Loop up/down but only in a limited range
	_rx += dY / 100 * 30;
	if( _rx > 90 ) _rx = 90; 
	if( _rx < -90 ) _rx = -90;
}
