// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
//
// Sample Application
// --------------------------------------
// Copyright (C) 2006-2011 Nicolas Schulz
//
//
// This sample source file is not covered by the EPL as the rest of the SDK
// and may be used without any restrictions. However, the EPL's disclaimer of
// warranty and liability shall be in effect for this file.
//
// *************************************************************************************************

#include "makenoise.h"
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "Horde3DUtils.h"
#include "bluenoise.h"
#include "Particle.h"

MakeNoise::
MakeNoise(void):
	noise_lengthscale(3),
	noise_gain(3),
	seed(0)
	{
		//noise_lengthscaleֱʾx,y,zٶȵƽ̶ȣֵԽ,ӼٶԽӽ
		//noise_gainֱʾx,y,zϵٶֵֵԽСӵnoiseԽС
		noise_lengthscale[0]=0.5;
		noise_gain[0]=0.3;
		noise_lengthscale[1]=0.5;
		noise_gain[1]=0.3;
		noise_lengthscale[2]=0.5;
		noise_gain[2]=0.3;

		initial_xmin = -0.5;initial_xmax = 0.5;
		initial_ymin = 7;initial_ymax = 8;
		initial_zmin = -0.5;initial_zmax = 0.5;
		sphere_centre1 = Vec3f(15 ,15, 15);
		sphere_radius1 = 0.0;
		sphere_centre2 = Vec3f(15, 15, 15);
		sphere_radius2 = 1.0;
		cube_centre = Vec3f(2, 2, 2);
		cube_length =2.0;
		
		d0 =0.5;
		d1 = 1.0;
		v0 = 1;
	}


bool MakeNoise::
seed_particles(std::vector< Particle > &particles, int totalnum) const
{
    if(t == 0)
	{
		Vec3f centre = sphere_centre2;
		float radius1 = sphere_radius1, radius2 = sphere_radius2, radiusSqr1 = radius1 * radius1, radiusSqr2 = radius2 * radius2;
		Particle p(0, 0.01);
		
		for (int i = 0; i < totalnum; i ++)
		{
			while (1)
			{
				p.Position = Vec3f(randhashf(seed ++, -radius2, radius2), randhashf(seed ++, -radius2, radius2), randhashf(seed ++, -radius2, radius2));
				float distance = mag2(p.Position);
				if (distance < radiusSqr1 || distance > radiusSqr2)
					continue;
				p.Position += centre;
				break;
			}
			particles.push_back(p);
		}
		return true;
	}
	else 
		return false;
	
	//bluenoise㷨ɵĳʼֲ
	//if(t==0){
	//	std::vector<Vec3f> x1;
	//	Particle p;
	//	p.node = 0;
	//	p.animTime = 0;
	//	bluenoise_sample(0.2f, Vec3f(initial_xmin, initial_ymin, initial_zmin), Vec3f(initial_xmax, initial_ymax, initial_zmax), x1);
	//	for (unsigned int i = 0; i < x1.size(); i ++)
	//	{
	//		p.x = x1[i];
	//		x.push_back(p);
	//	}
	//	return true;
	//}else
	//	return false; // no particles added after initial seeding at t=0

	//ʹwangtile㷨ɵĳʼֲ
	//if (t==0){
	//	LoadPoints(x,"ResultPoints.dat");
	//	return true;
	//}else

	//	return false;

}


float MakeNoise::
distance_and_normal(float x, float y, float z, Vec3f &normal) const
{
	float phi=y;
	Vec3f u(x-sphere_centre1[0], y-sphere_centre1[1], z-sphere_centre1[2]);
	float d=mag(u);
	if(d-sphere_radius1<phi){
		phi=d-sphere_radius1;
		normal=u/(d+1e-10);
	}else{
		normal=Vec3f(0,1,0);
	}
	return phi;
}

Vec3f MakeNoise::
normal(Vec3f u) const
{
	float d=mag(u);
	if (d >= 1e-10)
		return u/d;
	else return Vec3f(0,1,0);
}

void MakeNoise::
match_boundary(Vec3f &psi, float d, float d0, Vec3f normal) const
{
	if (d < 0) normal = - normal;
	float alpha = ramp(fabs(d)/d0);
	//if (alpha > 1)alpha = 1;
	//if (alpha < 0)alpha = 0;
	Vec3f tmp = (dot(normal,psi) * normal);
	//float tmp1 = mag(tmp);
	psi =alpha * psi + (1 - alpha) * tmp;
}

void MakeNoise::
match_sphere(float x, float y, float z, Vec3f &psi, float d0) const
{
	Vec3f u(x - sphere_centre1[0], y - sphere_centre1[1], z - sphere_centre1[2]);
	float dist = mag(u) - sphere_radius1;
	Vec3f n = normal(u);
	if (dist < 0) n = - n;
	float alpha = ramp(fabs(dist) / d0);
	Vec3f tmp = dot(n, psi) * n;
	psi = alpha * psi + (1 - alpha) * tmp;


	u = Vec3f(x - sphere_centre2[0], y - sphere_centre2[1], z - sphere_centre2[2]);
	dist = mag(u) - sphere_radius2;
	n = normal(u);
	if (dist < 0) n = -n;
	alpha = ramp(fabs(dist) / d0);
	tmp = dot(n, psi) * n;
	psi = alpha * psi + (1 - alpha) * tmp;

}

void MakeNoise::
match_cube(float x, float y, float z, Vec3f &psi, float d0) const
{
	Vec3f u(x - cube_centre[0], y - cube_centre[1], z - cube_centre[2]);
	Vec3f n = Vec3f(0, 0, -1);

	float dist = 1e10;
	if (fabs(cube_length - u[0]) < dist)
	{
		dist = fabs(cube_length - u[0]);
		n = Vec3f(1, 0, 0);
	}
	if (fabs(u[0] + cube_length) < dist)
	{
		dist = fabs(u[0] + cube_length);
		n = Vec3f(-1, 0, 0);
	}

	if (fabs(cube_length - u[1]) < dist)
	{
		dist = fabs(cube_length - u[1]);
		n = Vec3f(0, 1, 0);
	}

	if (fabs(u[1] + cube_length) < dist)
	{
		dist = fabs(u[1] + cube_length);
		n = Vec3f(0, -1, 0);
	}

	if (fabs(cube_length - u[2]) < dist)
	{
		dist = fabs(cube_length - u[2]);
		n = Vec3f(0, 0, 1);
	}

	if (fabs(u[2] + cube_length) < dist)
	{
		dist = fabs(u[2] + cube_length);
		n = Vec3f(0, 0, -1);
	}

	n = - n;
	match_boundary(psi, dist, d0, n);
}

void MakeNoise::
match_motion(float x, float y, float z, Vec3f &psi, float lengthscale) const
{
	Vec3f u(x - sphere_centre1[0], y - sphere_centre1[1], z - sphere_centre1[2]);
	float dist = mag(u);
	Vec3f n = normal(u);

	float d = dist - sphere_radius1;
	float alpha = ramp(fabs(d)/d0);
	psi =alpha * psi + (1 - alpha) * (dot(n,psi) * n);
	alpha = 1 / (((fabs(d))) + 1.0f);
	psi += cross(Vec3f(x,y,z),(v0 * (n)));
}


Vec3f MakeNoise::
potential(float x, float y, float z) const
{
	Vec3f psi(0,0,0);

	for(unsigned int i=0; i<noise_lengthscale.size(); ++i){
		float sx=x/noise_lengthscale[i];
		float sy=y/noise_lengthscale[i];
		float sz=z/noise_lengthscale[i];
		Vec3f psi_i(noise0(sx,sy,sz), noise1(sx,sy,sz), noise2(sx,sy,sz));
		//psi_i = psi_i + 0.2f * Vec3f(z, 0, -x);  //y᷽һ0.1ٶ
		//match_boundary(psi_i, d, noise_lengthscale[i], normal);
		match_sphere(x, y, z, psi_i, d0);
		// match_cube(x, y, z, psi_i, d0);
		//match_motion(x, y ,z, psi_i, d0);
		psi += noise_gain[i]*psi_i;
	}

	// add rising vortex rings
	//float ring_y=ring_speed*t;
	//while(ring_y>0){
	//   float ry=y-ring_y;
	//   float rr=std::sqrt(x*x+z*z);
	//   float rmag=ring_magnitude/(sqr(rr-ring_radius)+sqr(rr+ring_radius)+sqr(ry)+ring_falloff);
	//   Vec3f rpsi=rmag*Vec3f(z, 0, -x);
	//   match_boundary(rpsi, d, ring_radius, normal);
	//   psi+=rpsi;
	//   ring_y-=ring_speed/rings_per_second;
	//}

	return psi;
}

void MakeNoise::
advance_time(float dt)
{
	t+=dt;
	noise.set_time(0.5f*noise_gain[0]/noise_lengthscale[0]*t);
}

