//Potential.cpp 
#include "Potential.h"
#include "interp.h"
#include <cmath>
extern FILE *fin, *fout;

Potential::Potential(int x, int y, int z) : gridX(x), gridY(y), gridZ(z) 
{
}

void Potential::SetGoal(int xa, int xb, int ya, int yb, int za, int zb)
{
	m_GoalXa = xa;
	m_GoalXb = xb;
	m_GoalYa = ya;
	m_GoalYb = yb;
	m_GoalZa = za;
	m_GoalZb = zb;
}

void Potential::SetSpeedRange(float _minSpeed, float _maxSpeed)
{
	maxSpeed = _maxSpeed;
	minSpeed = _minSpeed;
}

void Potential::SetDensityRange(float _minDensity, float _maxDensity)
{
	maxDensity = _maxDensity;
	minDensity = _minDensity;
}

void Potential::SetLamdaWeights(float _lamda, float _w1, float _w2, float _w3)
{
	lamda = _lamda;
	w1 = _w1;
	w2 = _w2;
	w3 = _w3;
}

void Potential::ComputeFaceData(void)
{
	int x, y, z;
	float speedT = minSpeed, speedV;
	for(x = 0; x < gridX; x ++)
		for(y = 0; y < gridY; y ++)
			for (z = 0; z < gridZ; z ++)
			{
				speedV = mag(get_DensityGradient(Vec3f(x, y, z)));
				if (x + 1 < gridX)
				{
					//speedV =  averagev[x + 1][y][z][0];
					if(speedV > speedT) speedV = speedT;
					if(density[x + 1][y][z] <= minDensity)
					{
						m_gridData[x][y][z].faceData[0].speed = speedT;
					}
					else if(density[x + 1][y][z] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[0].speed = speedV;
					}
					else
					{
						m_gridData[x][y][z].faceData[0].speed = speedT + (density[x + 1][y][z] - minDensity)
							/(maxDensity - minDensity) * (speedV-speedT);
					}
				}
				if (y + 1 < gridY)
				{
					//speedV =  abs(averagev[x][y + 1][z][1]);
					if(speedV > speedT) speedV = speedT;
					if(density[x][y + 1][z] <= minDensity)
					{
						m_gridData[x][y][z].faceData[1].speed = speedT;
					}
					else if(density[x][y + 1][z] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[1].speed = speedV;
					}
					else
					{
						m_gridData[x][y][z].faceData[1].speed = speedT + (density[x][y + 1][z] - minDensity)
							/(maxDensity - minDensity) * (speedV-speedT);
					}
				}
				if (z + 1 < gridZ)
				{
					//speedV =  abs(averagev[x][y][z + 1][2]);
					if(speedV > speedT) speedV = speedT;
					if(density[x][y][z + 1] <= minDensity)
					{
						m_gridData[x][y][z].faceData[2].speed = speedT;
					}
					else if(density[x][y][z + 1] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[2].speed = speedV;
					}
					else
					{
						m_gridData[x][y][z].faceData[2].speed = speedT + (density[x][y][z + 1] - minDensity)
							/(maxDensity - minDensity) * (speedV-speedT);
					}
				}
				if (x - 1 >= 0)
				{
					//speedV =  -averagev[x - 1][y][z][0];
					if(speedV < 0) speedV = 0;
					if(speedV > speedT) speedV = speedT;
					if(density[x - 1][y][z] <= minDensity)
					{
						m_gridData[x][y][z].faceData[3].speed = speedT;
					}
					else if(density[x - 1][y][z] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[3].speed = speedV;
					}
					else
					{ 
						m_gridData[x][y][z].faceData[3].speed = speedT + (density[x - 1][y][z] - minDensity) 
							/ (maxDensity - minDensity) * (speedV - speedT);
					}
				}
				if (y - 1 >= 0)
				{
					//speedV =  -averagev[x][y - 1][z][1];
					if(speedV < 0) speedV = 0;
					if(speedV > speedT) speedV = speedT;
					if(density[x][y - 1][z] <= minDensity)
					{
						m_gridData[x][y][z].faceData[4].speed = speedT;
					}
					else if(density[x][y - 1][z] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[4].speed = speedV;
					}
					else
					{ 
						m_gridData[x][y][z].faceData[4].speed = speedT + (density[x][y - 1][z] - minDensity) 
							/ (maxDensity - minDensity) * (speedV - speedT);
					}
				}
				if(z - 1 >= 0)
				{
					//speedV =  -averagev[x][y][z - 1][2];
					if(speedV < 0) speedV = 0;
					if(speedV > speedT) speedV = speedT;
					if(density[x][y][z - 1] <= minDensity)
					{
						m_gridData[x][y][z].faceData[5].speed = speedT;
					}
					else if(density[x][y][z - 1] >= maxDensity)
					{
						m_gridData[x][y][z].faceData[5].speed = speedV;
					}
					else
					{ 
						m_gridData[x][y][z].faceData[5].speed = speedT + (density[x][y][z - 1] - minDensity) 
							/ (maxDensity - minDensity) * (speedV - speedT);
					}
				}
			}

	for(x = 0; x < gridX; x ++)
		for (y = 0; y < gridY; y ++) 
			for( z = 0; z < gridZ; z ++)
			{
				if(x + 1 < gridX)
				{	
					if(m_gridData[x][y][z].faceData[0].speed <= 0) 
						m_gridData[x][y][z].faceData[0].speed = MININT;
					m_gridData[x][y][z].faceData[0].cost = (w1 * m_gridData[x][y][z].faceData[0].speed + w2 +
						w3 * m_gridData[x + 1][y][z].discomfort) / m_gridData[x][y][z].faceData[0].speed;
				}
				if(y + 1 < gridY)
				{	
					if(m_gridData[x][y][z].faceData[1].speed <= 0) 
						m_gridData[x][y][z].faceData[1].speed = MININT;
					m_gridData[x][y][z].faceData[1].cost = (w1 * m_gridData[x][y][z].faceData[1].speed + w2 +
						w3 * m_gridData[x][y + 1][z].discomfort) / m_gridData[x][y][z].faceData[1].speed;
				}
				if(z + 1 < gridZ)
				{	
					if(m_gridData[x][y][z].faceData[2].speed <= 0) 
						m_gridData[x][y][z].faceData[2].speed = MININT;
					m_gridData[x][y][z].faceData[2].cost = (w1 * m_gridData[x][y][z].faceData[2].speed + w2 +
						w3 * m_gridData[x][y][z + 1].discomfort) / m_gridData[x][y][z].faceData[2].speed;
				}

				if(x - 1 >= 0)
				{
					if(m_gridData[x][y][z].faceData[3].speed <= 0) 
						m_gridData[x][y][z].faceData[3].speed = MININT;
					m_gridData[x][y][z].faceData[3].cost = (w1 * m_gridData[x][y][z].faceData[3].speed + w2 + 
						w3 * m_gridData[x - 1][y][z].discomfort) / m_gridData[x][y][z].faceData[3].speed;
				}
				if(y - 1 >= 0)
				{
					if(m_gridData[x][y][z].faceData[4].speed <= 0) 
						m_gridData[x][y][z].faceData[4].speed = MININT;
					m_gridData[x][y][z].faceData[4].cost = (w1 * m_gridData[x][y][z].faceData[4].speed + w2 + 
						w3 * m_gridData[x][y - 1][z].discomfort) / m_gridData[x][y][z].faceData[4].speed;
				}
				if(z - 1 >= 0)
				{
					if(m_gridData[x][y][z].faceData[5].speed <= 0) 
						m_gridData[x][y][z].faceData[5].speed = MININT;
					m_gridData[x][y][z].faceData[5].cost = (w1 * m_gridData[x][y][z].faceData[5].speed + w2 + 
						w3 * m_gridData[x][y][z - 1].discomfort) / m_gridData[x][y][z].faceData[5].speed;
				}
			}
}

//processڵtargetΪCANDIDATE״̬CandidateУPotential
void Potential::process(int x, int y, int z, priority_queue<Node, vector<Node>, cmp>& Candidate)
{
	if (m_gridData[x][y][z].FMMFlag != UNKNOWN)
		return;

	CalcPotential(x, y, z);

	m_gridData[x][y][z].FMMFlag = CANDIDATE;
	Node tmp;
	tmp.x = x;
	tmp.y = y;
	tmp.z = z;
	tmp.value = m_gridData[x][y][z].potential;
	Candidate.push(tmp);
}
void Potential::FMM()
{
	int tmpgridX = gridX - 2, tmpgridY = gridY - 2, tmpgridZ = gridZ - 2;
	int totalGrids = (tmpgridX - 2) * (tmpgridY - 2) * (tmpgridZ - 2);//
	priority_queue<Node, vector<Node>, cmp> Candidate;
	int x, y, z;
	int KnownIndex = 0;


	//ʼ еPotentialΪ״̬ΪUNKOWN
	for(x = 0; x < gridX; x ++)
		for (y = 0; y < gridY; y ++)
			for(z = 0; z < gridZ; z ++)
			{
				m_gridData[x][y][z].FMMFlag = UNKNOWN;
				m_gridData[x][y][z].potential = MAXINT;
			}

	//ĿPotentialΪ0״̬ΪKNOWNʼCandidateѡ
	for(x = m_GoalXa; x <= m_GoalXb; x ++)
		for (y = m_GoalYa; y <= m_GoalYb; y ++)
			for(z = m_GoalZa; z <= m_GoalZb; z ++)
			{
				m_gridData[x][y][z].FMMFlag = KNOWN;
				m_gridData[x][y][z].potential = 0;
				KnownIndex ++;
				if (x + 1 < tmpgridX) process(x + 1, y, z, Candidate);
				if (y + 1 < tmpgridY) process(x, y + 1, z, Candidate);
				if (z + 1 < tmpgridZ) process(x, y, z + 1, Candidate);

				if (x - 1 >= 2) process(x - 1, y, z, Candidate);
				if (y - 1 >= 2) process(x, y - 1, z, Candidate);
				if (z - 1 >= 2) process(x, y, z - 1, Candidate);
			}

	/*KnownIndexδʱѭȰCandidatePotentialֵСĽڵΪKNOWNýڵھӣ㣩
	  CandidateвPotentialʽ11*/
	while (KnownIndex < totalGrids)
	{
		//ȡ׵ԪأڵֵΪKNOWN
		Node tmp = Candidate.top();
		Candidate.pop();
		x = tmp.x;
		y = tmp.y;
		z = tmp.z;
		//ǰڵСֵȶдŵֵͬ˵
		//ýڵѾpopˣѾõСֵٴ
		//if (tmp.value != m_gridData[x][y][z].potential)
			//continue;

		m_gridData[x][y][z].FMMFlag = KNOWN;
		KnownIndex ++;
		if (x + 1 < tmpgridX) process(x + 1, y, z, Candidate);
		if (y + 1 < tmpgridY) process(x, y + 1, z, Candidate);
		if (z + 1 < tmpgridZ) process(x, y, z + 1, Candidate);

		if (x - 1 >= 2) process(x - 1, y, z, Candidate);
		if (y - 1 >= 2) process(x, y - 1, z, Candidate);
		if (z - 1 >= 2) process(x, y, z - 1, Candidate);
	}

	return;
}
void Potential::solve(float &x, float a, float b, float c, float d)
{
	float A = d + b;
	float B = - 2 * a * d - 2 * c * b;
	float C = a * a * d + c * c * b - b * d;
	float D = B * B - 4 * A * C;
	if(D < 0) D = 0;
	x = (- B + sqrt(D)) / (2.0 * A);
}
void Potential::CalcPotential(int x, int y, int z)
{
	float totalmin = MAXINT;
	bool minx = false;
	float a, b = 0, c, d = 0, e ,f = 0;
	if (x < gridX - 1 && m_gridData[x + 1][y][z].potential + m_gridData[x][y][z].faceData[0].cost < totalmin)
	{
		a = m_gridData[x + 1][y][z].potential;
		b = m_gridData[x][y][z].faceData[0].cost;
		totalmin = a + b;
		minx = true; 
	}
	if (x > 0 && m_gridData[x - 1][y][z].potential + m_gridData[x][y][z].faceData[3].cost < totalmin)
	{
		a = m_gridData[x - 1][y][z].potential;
		b = m_gridData[x][y][z].faceData[3].cost;
		minx = true;
	}

	totalmin = MAXINT;
	bool miny = false;
	if (y < gridY - 1 && m_gridData[x][y + 1][z].potential + m_gridData[x][y][z].faceData[1].cost < totalmin)
	{
		c = m_gridData[x][y + 1][z].potential;
		d = m_gridData[x][y][z].faceData[1].cost;
		totalmin = c + d;
		miny = true;
	}
	if (y > 0 && m_gridData[x][y - 1][z].potential + m_gridData[x][y][z].faceData[4].cost < totalmin)
	{
		c = m_gridData[x][y - 1][z].potential;
		d = m_gridData[x][y][z].faceData[4].cost;
		miny = true;
	}

	totalmin = MAXINT;
	bool minz = false;
	if (z < gridZ - 1 && m_gridData[x][y][z + 1].potential + m_gridData[x][y][z].faceData[2].cost < totalmin)
	{
		e = m_gridData[x][y][z + 1].potential;
		f = m_gridData[x][y][z].faceData[2].cost;
		totalmin = e + f;
		minz = true;
	}
	if (z > 0 && m_gridData[x][y][z - 1].potential + m_gridData[x][y][z].faceData[5].cost < totalmin)
	{
		e = m_gridData[x][y][z - 1].potential;
		f = m_gridData[x][y][z].faceData[5].cost;
		minz = true;
	}

	//㹫ʽ11.㷨Ҫ㲻֪дôuglyĳȫelse if
	if (!minx && !miny)
		m_gridData[x][y][z].potential = e + f;
	else if (!minx && !minz)
		m_gridData[x][y][z].potential = c + d;
	else if (!miny && !minz)
		m_gridData[x][y][z].potential = a + b;
	else 
	{
		//ƽü㡣
		b = b * b;
		d = d * d;
		f = f * f;

		if (!minx)
			solve(m_gridData[x][y][z].potential, c, d, e, f);
		else if (!miny)
			solve(m_gridData[x][y][z].potential, a, b, e, f);
		else if (!minz)
			solve(m_gridData[x][y][z].potential, a, b, c, d);
		else
		{
			//ߵһ˵·̫ˡ
			float A = d * f + b * f + b * d;
			float B = - 2 * d * f * a - 2 * b * f * c - 2 * b * d * e;
			float C = a * a * d * f + c * c * b * f + e * e * b * d - b * d * f;
			float D = B * B - 4 * A * C;
			if (D < 0) D = 0;
			m_gridData[x][y][z].potential = (- B + sqrt(D)) / (2.0 * A);
		}

		//if (!minx && !miny && !minz)
		//{
		//	minx = 0;
		//}
	}
}
void Potential::addDensity(Vec3f position, Vec3f velocity)
{
	float x = position[0] / GRIDL;
	float y = position[1] / GRIDL;
	float z = position[2] / GRIDL;
	int i = floor(x + 0.5);
	int j = floor(y + 0.5);
	int k = floor(z + 0.5);
	//Ӧsection 4.1㷨
	if (i > gridX || j > gridY || k > gridZ || i < 0 || j < 0 || k < 0) return;

	float deltax = x - i + 0.5 * GRIDL;
	float deltay = y - j + 0.5 * GRIDL;
	float deltaz = z - k + 0.5 * GRIDL;
	float tmp;
	if (i > 0 && j > 0 && k < gridZ)  //density A
	{
		tmp = min(1 - deltax, 1 - deltay, deltaz);
		tmp = pow(tmp, lamda);
		density[i - 1][j - 1][k] += tmp;
		averagev[i - 1][j - 1][k] += velocity * tmp;
	}
	if (i < gridX && j > 0 && k < gridZ) //density B
	{
		tmp = min(deltax, 1 - deltay, deltaz);
		tmp = pow(tmp, lamda);
		density[i][j - 1][k] += tmp;
		averagev[i][j - 1][k] += velocity * tmp;
	}
	if (i < gridX && j < gridY && k < gridZ) //density C
	{
		tmp = min(deltax, deltay, deltaz);
		tmp = pow(tmp, lamda);
		density[i][j][k] += tmp;
		averagev[i][j][k] += velocity * tmp;
	}
	if (i > 0 && j < gridY && k < gridZ) //density D
	{
		tmp = min(1 - deltax, deltay, deltaz);
		tmp = pow(tmp, lamda);
		density[i - 1][j][k] += tmp;
		averagev[i - 1][j][k] += velocity * tmp;
	}

	if (i > 0 && j > 0 && k > 0)  //density E
	{
		tmp = min(1 - deltax, 1 - deltay, 1 - deltaz);
		tmp = pow(tmp, lamda);
		density[i - 1][j - 1][k - 1] += tmp;
		averagev[i - 1][j - 1][k - 1] += velocity * tmp;
	}
	if (i < gridX && j > 0 && k > 0) //density F
	{
		tmp = min(deltax, 1 - deltay, 1 - deltaz);
		tmp = pow(tmp, lamda);
		density[i][j - 1][k - 1] += tmp;
		averagev[i][j - 1][k - 1] += velocity * tmp;
	}
	if (i < gridX && j < gridY && k > 0) //density G
	{
		tmp = min(deltax, deltay, 1 - deltaz);
		tmp = pow(tmp, lamda);
		density[i][j][k - 1] += tmp;
		averagev[i][j][k - 1] += velocity * tmp;
	}
	if (i > 0 && j < gridY && k > 0) //density H
	{
		tmp = min(1 - deltax, deltay, 1 - deltaz);
		tmp = pow(tmp, lamda);
		density[i - 1][j][k - 1] += tmp;
		averagev[i - 1][j][k - 1] += velocity * tmp;
	}
}
Vec3f Potential::GetVelocity(Vec3f position)
{
	//ֵgradient
	Vec3f gradient = get_PotentialGradient(position);
	gradient = normalized(-gradient);

	//i,j,kʾposition
	int i = floor(position[0] / GRIDL);
	int j = floor(position[1] / GRIDL);
	int k = floor(position[2] / GRIDL);
	if (m_gridData[i][j][k].potential == 0) 
		return(Vec3f(0, 0, 0));
	//speed
	//x,y,zʾpositiongradientλ
	int x = floor(position[0] / GRIDL + gradient[0] / 2 + 0.5);
	int y = floor(position[1] / GRIDL + gradient[1] / 2 + 0.5);
	int z = floor(position[2] / GRIDL + gradient[2] / 2 + 0.5);
	float speed = m_gridData[i][j][k].faceData[(-3) * (x - i) + 3].speed * abs(gradient[0]) + 
		m_gridData[i][j][k].faceData[(-3) * (y - j) + 4].speed * abs(gradient[1]) +
		m_gridData[i][j][k].faceData[(-3) * (z - k) + 5].speed * abs(gradient[2]);
	Vec3f velocity = gradient * speed;
	return velocity;
}
Vec3f Potential::get_PotentialGradient(Vec3f position)
{
	int i,j,k;
	float fx,fy,fz;
	Vec3f gradient;

	get_barycentric(position[0], i, fx, 0, gridX);
	get_barycentric(position[1], j, fy, 0, gridY);
	get_barycentric(position[2], k, fz, 0, gridZ);
	//dx
	float v000 = (m_gridData[i+1][j][k].potential - m_gridData[i-1][j][k].potential) / GRIDL * 0.5f;
	float v001 = (m_gridData[i+1][j][k+1].potential - m_gridData[i-1][j][k+1].potential) / GRIDL * 0.5f;
	float v010 = (m_gridData[i+1][j+1][k].potential - m_gridData[i-1][j+1][k].potential) / GRIDL*0.5f;
	float v011 = (m_gridData[i+1][j+1][k+1].potential - m_gridData[i-1][j+1][k+1].potential) / GRIDL * 0.5f;
	float v100 = (m_gridData[i+2][j][k].potential - m_gridData[i][j][k].potential) / GRIDL * 0.5f;
	float v101 = (m_gridData[i+2][j][k+1].potential-m_gridData[i][j][k+1].potential) / GRIDL * 0.5f;
	float v110 = (m_gridData[i+2][j+1][k].potential-m_gridData[i][j+1][k].potential) / GRIDL * 0.5f;
	float v111 = (m_gridData[i+2][j+1][k+1].potential - m_gridData[i][j+1][k+1].potential) / GRIDL * 0.5f;
	gradient[0]=trilerp(v000, v100, v010, v110, v001, v101, v011, v111, fx, fy, fz);
	//dy
	v000 = (m_gridData[i][j+1][k].potential-m_gridData[i][j-1][k].potential)/GRIDL*0.5f;
	v001 = (m_gridData[i][j+1][k+1].potential-m_gridData[i][j-1][k+1].potential)/GRIDL*0.5f;
	v010 = (m_gridData[i][j+2][k].potential-m_gridData[i][j][k].potential)/GRIDL*0.5f;
	v011 = (m_gridData[i][j+2][k+1].potential-m_gridData[i][j][k+1].potential)/GRIDL*0.5f;
	v100 = (m_gridData[i+1][j+1][k].potential-m_gridData[i+1][j-1][k].potential)/GRIDL*0.5f;
	v101 = (m_gridData[i+1][j+1][k+1].potential-m_gridData[i+1][j-1][k+1].potential)/GRIDL*0.5f;
	v110 = (m_gridData[i+1][j+2][k].potential-m_gridData[i+1][j][k].potential)/GRIDL*0.5f;
	v111 = (m_gridData[i+1][j+2][k+1].potential-m_gridData[i+1][j][k+1].potential)/GRIDL*0.5f;
	gradient[1]=trilerp(v000, v100,v010, v110, v001, v101,v011, v111,fx, fy, fz);
	//dz
	v000 = (m_gridData[i][j][k+1].potential-m_gridData[i][j][k-1].potential)/GRIDL*0.5f;
	v001 = (m_gridData[i][j][k+2].potential-m_gridData[i][j][k].potential)/GRIDL*0.5f;
	v010 = (m_gridData[i][j+1][k+1].potential-m_gridData[i][j+1][k-1].potential)/GRIDL*0.5f;
	v011 = (m_gridData[i][j+1][k+2].potential-m_gridData[i][j+1][k].potential)/GRIDL*0.5f;
	v100 = (m_gridData[i+1][j][k+1].potential-m_gridData[i+1][j][k-1].potential)/GRIDL*0.5f;
	v101 = (m_gridData[i+1][j][k+2].potential-m_gridData[i+1][j][k].potential)/GRIDL*0.5f;
	v110 = (m_gridData[i+1][j+1][k+1].potential-m_gridData[i+1][j+1][k-1].potential)/GRIDL*0.5f;
	v111 = (m_gridData[i+1][j+1][k+2].potential-m_gridData[i+1][j+1][k].potential)/GRIDL*0.5f;
	gradient[2]=trilerp(v000, v100,v010, v110, v001, v101,v011, v111,fx, fy, fz);

	return gradient;
}

Vec3f Potential::get_DensityGradient(Vec3f position)
{
	int i,j,k;
	float fx,fy,fz;
	Vec3f gradient;

	get_barycentric(position[0], i, fx, 0, gridX);
	get_barycentric(position[1], j, fy, 0, gridY);
	get_barycentric(position[2], k, fz, 0, gridZ);
	//dx
	float v000 = (density[i+1][j][k] - density[i-1][j][k])* 0.5f;
	float v001 = (density[i+1][j][k+1] - density[i-1][j][k+1])* 0.5f;
	float v010 = (density[i+1][j+1][k] - density[i-1][j+1][k])*0.5f;
	float v011 = (density[i+1][j+1][k+1] - density[i-1][j+1][k+1])* 0.5f;
	float v100 = (density[i+2][j][k] - density[i][j][k])* 0.5f;
	float v101 = (density[i+2][j][k+1]-density[i][j][k+1])* 0.5f;
	float v110 = (density[i+2][j+1][k]-density[i][j+1][k])* 0.5f;
	float v111 = (density[i+2][j+1][k+1] - density[i][j+1][k+1])* 0.5f;
	gradient[0]=trilerp(v000, v100, v010, v110, v001, v101, v011, v111, fx, fy, fz);
	//dy
	v000 = (density[i][j+1][k]-density[i][j-1][k])*0.5f;
	v001 = (density[i][j+1][k+1]-density[i][j-1][k+1])*0.5f;
	v010 = (density[i][j+2][k]-density[i][j][k])*0.5f;
	v011 = (density[i][j+2][k+1]-density[i][j][k+1])*0.5f;
	v100 = (density[i+1][j+1][k]-density[i+1][j-1][k])*0.5f;
	v101 = (density[i+1][j+1][k+1]-density[i+1][j-1][k+1])*0.5f;
	v110 = (density[i+1][j+2][k]-density[i+1][j][k])*0.5f;
	v111 = (density[i+1][j+2][k+1]-density[i+1][j][k+1])*0.5f;
	gradient[1]=trilerp(v000, v100,v010, v110, v001, v101,v011, v111,fx, fy, fz);
	//dz
	v000 = (density[i][j][k+1]-density[i][j][k-1])*0.5f;
	v001 = (density[i][j][k+2]-density[i][j][k])*0.5f;
	v010 = (density[i][j+1][k+1]-density[i][j+1][k-1])*0.5f;
	v011 = (density[i][j+1][k+2]-density[i][j+1][k])*0.5f;
	v100 = (density[i+1][j][k+1]-density[i+1][j][k-1])*0.5f;
	v101 = (density[i+1][j][k+2]-density[i+1][j][k])*0.5f;
	v110 = (density[i+1][j+1][k+1]-density[i+1][j+1][k-1])*0.5f;
	v111 = (density[i+1][j+1][k+2]-density[i+1][j+1][k])*0.5f;
	gradient[2]=trilerp(v000, v100,v010, v110, v001, v101,v011, v111,fx, fy, fz);

	return gradient;
}



bool Potential::addObj(const char* filename)
{
	objloader model;
	if (!model.ReadObjFromFile(filename))
	{
		fprintf(fout, "Cannot find the %s, please input again.\n", filename);
		return 0;
	}

	//ʼdistance
	int i, j, k;
	for (i = 0 ;i < gridX; i ++)
		for (j = 0; j < gridY; j ++)
			for (k = 0; k < gridZ; k ++)
				distance[i][j][k] = MAXINT;

	int near_i,near_j,near_k;
	int i_min,i_max,j_min,j_max,k_min,k_max;
	Vec3f point_a,point_b,point_c;
	int index_a,index_b,index_c;
	for (i = 0; i < model.fNodeNum; i ++)
	{
		point_a = model.vNodeList[model.fNodeList[i].nodes[0].vNum];
		point_b = model.vNodeList[model.fNodeList[i].nodes[1].vNum];
		point_c = model.vNodeList[model.fNodeList[i].nodes[2].vNum];

		i_max=max((point_a[0])/GRIDL,(point_b[0])/GRIDL,(point_c[0])/GRIDL);
		i_min=min((point_a[0])/GRIDL,(point_b[0])/GRIDL,(point_c[0])/GRIDL);

		j_max=max((point_a[1])/GRIDL,(point_b[1])/GRIDL,(point_c[1])/GRIDL);
		j_min=min((point_a[1])/GRIDL,(point_b[1])/GRIDL,(point_c[1])/GRIDL);

		k_max=max((point_a[2])/GRIDL,(point_b[2])/GRIDL,(point_c[2])/GRIDL);
		k_min=min((point_a[2])/GRIDL,(point_b[2])/GRIDL,(point_c[2])/GRIDL);

		for (near_k = k_min - 3; near_k < k_max + 3; near_k ++)
			for (near_j = j_min - 3; near_j < j_max + 3; near_j ++)
				for (near_i = i_min - 3; near_i < i_max + 3; near_i ++)
				if (near_i >= 0 && near_i < gridX && near_j >= 0 && near_j < gridY && near_k >= 0 && near_k < gridZ)
				{
					float dis;
					Vec3f grid_point((near_i + 0.5) * GRIDL, (near_j + 0.5) * GRIDL, (near_k + 0.5) * GRIDL);

					dis = ComputeDistance(model, i, grid_point);
					if (fabs(dis) < fabs(distance[near_i][near_j][near_k]))
						distance[near_i][near_j][near_k] = dis;
				}
	}

	Floodfill(gridX - 1, gridY - 1, gridZ - 1);
	for (j = 0 ;j < gridY; j ++)
	{
		for (i = 0; i < gridX; i ++)

			for (k = 0; k < gridZ; k ++)
				fprintf(fout, "%f ", distance[i][j][k]);
		fprintf(fout, "\n");
	}
		//distance㣨ۼӣdiscomfortֵ
		ComputeDiscomfort();
}

void Potential::ComputeDiscomfort(void)
{
	int i, j, k;
	for (i = 0; i < gridX; i++)
		for (j = 0; j < gridY; j ++)
			for (k = 0; k < gridZ; k ++)
			{
				if (distance[i][j][k] > 0) //˵meshڲ
				{
					m_gridData[i][j][k].discomfort += DISVALUE;
				}
				//else if (distance[i][j][k] != -MAXINT) //˵meshⲿmesh߽
				//{
				//	//ֻҪڵһֵ˵meshռݣDiscomfortӦóɽϴֵ
				//	if (   (i > 0 && distance[i - 1][j][k] > 0) || (i < gridX - 1 && distance[i + 1][j][k] > 0)
				//		|| (j > 0 && distance[i][j - 1][k] > 0) || (j < gridY - 1 && distance[i][j + 1][k] > 0)
				//		|| (k > 0 && distance[i][j][k - 1] > 0) || (k < gridZ - 1 && distance[i][j][k + 1] > 0))
				//		m_gridData[i][j][k].discomfort += DISVALUE;
				//}
			}

}

void Potential::Floodfill(int i, int j, int k)
{
	if (distance[i][j][k] != MAXINT)
		return;
	distance[i][j][k] = - MAXINT;
	if (i > 0)
		Floodfill(i - 1, j, k);
	if (i < gridX - 1)
		Floodfill(i + 1, j, k);
	if (j > 0)
		Floodfill(i, j - 1, k);
	if (j < gridY - 1)
		Floodfill(i, j + 1, k);
	if (k > 0)
		Floodfill(i, j, k - 1);
	if (k < gridZ - 1)
		Floodfill(i, j, k + 1);
}
//objģͶƬƬƬ㣬ǰ񶥵
float Potential::ComputeDistance(objloader model, int trangle_index, Vec3f point)
{
	Vec3f normal_a = model.vnNodeList[model.fNodeList[trangle_index].nodes[0].vnNum];
	Vec3f normal_b = model.vnNodeList[model.fNodeList[trangle_index].nodes[1].vnNum];
	Vec3f normal_c = model.vnNodeList[model.fNodeList[trangle_index].nodes[2].vnNum];

	Vec3f point_a = model.vNodeList[model.fNodeList[trangle_index].nodes[0].vNum];
	Vec3f point_b = model.vNodeList[model.fNodeList[trangle_index].nodes[1].vNum];
	Vec3f point_c = model.vNodeList[model.fNodeList[trangle_index].nodes[2].vNum];

	float dis,t;
	Vec3f facenormal,intersection_point,sign_normal; //ķ򣬽
	float edge_A,edge_B,edge_C;              //߳
	float dis_of_intersectionpoint_to_edgeA,dis_of_intersectionpoint_to_edgeB,dis_of_intersectionpoint_to_edgeC;    //㵽ߵľ
	Vec3f  Stagnation_A,Stagnation_B,Stagnation_C;       //㵽ÿߵĽŵ
	float dis_to_pointA,dis_to_pointB,dis_to_pointC;    //㵽ÿľ
	float trangle_area,sum_area;
	bool flag_edgeA,flag_edgeB,flag_edgeC;
	float temp_disA,temp_disB,temp_disC;
	float sign_A,sign_B,sign_C,sign_edge_A,sign_edge_B,sign_edge_C;
	//ķ
	facenormal = model.fNodeList[trangle_index].fnormal;

	t = dot(facenormal, point_a - point) / mag2(facenormal);
	intersection_point = facenormal * t + point;

	edge_A = mag(point_b - point_c);
	edge_B = mag(point_a - point_c);
	edge_C = mag(point_a - point_b);

	t = dot(point_b - point_c, point - point_b) 
		/ mag2(point_b - point_c);
	Stagnation_A = (point_b - point_c) * t + point_b;

	t = dot(point_a - point_c, point - point_a) 
		/ mag2(point_a - point_c);
	Stagnation_B = (point_a-point_c) * t + point_a;

	t = dot(point_a - point_b, point - point_b) 
		/ mag2(point_a - point_b);
	Stagnation_C = (point_a - point_b) * t + point_b;

	dis_of_intersectionpoint_to_edgeA = mag(intersection_point - Stagnation_A);
	dis_of_intersectionpoint_to_edgeB = mag(intersection_point - Stagnation_B);
	dis_of_intersectionpoint_to_edgeC = mag(intersection_point - Stagnation_C);

	trangle_area = sqrt((edge_A + edge_B + edge_C) * (edge_A + edge_B - edge_C)
					* (edge_B + edge_C - edge_A) * (edge_A + edge_C - edge_B)) / 4.0;
	sum_area = (edge_A * dis_of_intersectionpoint_to_edgeA + 
				edge_B * dis_of_intersectionpoint_to_edgeB + 
				edge_C * dis_of_intersectionpoint_to_edgeC) / 2.0;

	dis_to_pointA = mag(point - point_a);
	dis_to_pointB = mag(point - point_b);
	dis_to_pointC = mag(point - point_c);

	//ڶA
	sign_normal = normal_a;
	sign_A = inside_or_outside(point - point_a,sign_normal);
	//ڶB
	sign_normal = normal_b;
	sign_B = inside_or_outside(point - point_b, sign_normal);
	//ڶC
	sign_normal = normal_c;
	sign_C = inside_or_outside(point - point_c, sign_normal);
	//ڱA
	sign_normal = normalized(normal_b + normal_c);
	sign_edge_A = inside_or_outside(point - Stagnation_A, sign_normal);
	//ڱB
	sign_normal = normalized(normal_a + normal_c);
	sign_edge_B = inside_or_outside(point - Stagnation_B, sign_normal);
	//ڱC
	sign_normal =normalized(normal_b + normal_a);
	sign_edge_C = inside_or_outside(point - Stagnation_C, sign_normal);

	if (((trangle_area - sum_area) <= MININT) && 
		((trangle_area-sum_area) >= - MININT))
	{
		dis = mag(point - intersection_point);
		//㲻ڱߺͶ
		if (dis_of_intersectionpoint_to_edgeA * dis_of_intersectionpoint_to_edgeB 
			* dis_of_intersectionpoint_to_edgeC > MININT)
		{
			dis = dis * inside_or_outside(point - intersection_point, facenormal);
			return dis;
		}
		//ڶC
		if ((dis_of_intersectionpoint_to_edgeA <= MININT) &&
			(dis_of_intersectionpoint_to_edgeA >= - MININT) && 
			(dis_of_intersectionpoint_to_edgeB <= MININT) && 
			(dis_of_intersectionpoint_to_edgeB >= - MININT))
		{
			dis = dis * sign_C;
			return dis;
		}
		//ڶA
		if ((dis_of_intersectionpoint_to_edgeB <= MININT) &&
			(dis_of_intersectionpoint_to_edgeB >= - MININT) &&
			(dis_of_intersectionpoint_to_edgeC <= MININT) &&
			(dis_of_intersectionpoint_to_edgeC >= - MININT))
		{
			dis = dis * sign_A;
			return dis;
		}
		//ڶB
		if ((dis_of_intersectionpoint_to_edgeA <= MININT) &&
			(dis_of_intersectionpoint_to_edgeA >= - MININT) &&
			(dis_of_intersectionpoint_to_edgeC <= MININT) &&
			(dis_of_intersectionpoint_to_edgeC >= - MININT))
		{
			dis = dis * sign_B;
			return dis;
		}
		//ڱA
		if ((dis_of_intersectionpoint_to_edgeA <= MININT) &&
			(dis_of_intersectionpoint_to_edgeA >= - MININT))
		{
			dis = dis * sign_edge_A;
			return dis;
		}
		//ڱB
		if ((dis_of_intersectionpoint_to_edgeB <= MININT) &&
			(dis_of_intersectionpoint_to_edgeB >= - MININT))
		{
			dis = dis * sign_edge_B;
			return dis;
		}
		//ڱC
		if ((dis_of_intersectionpoint_to_edgeC <= MININT) &&
			(dis_of_intersectionpoint_to_edgeC >= - MININT))
		{
			dis = dis * sign_edge_C;
			return dis;
		} 
	}
	else
	{
		flag_edgeA = Is_point_on_segment(Stagnation_A, point_b, point_c);
		flag_edgeB = Is_point_on_segment(Stagnation_B, point_a, point_c);
		flag_edgeC = Is_point_on_segment(Stagnation_C, point_a, point_b);
		temp_disA = temp_disB = temp_disC = 1000.0;
		if (flag_edgeA)
			temp_disA = mag(point - Stagnation_A);
		if (flag_edgeB)
			temp_disB = mag(point - Stagnation_B);
		if (flag_edgeC)
			temp_disC = mag(point - Stagnation_C);

		dis = min(temp_disA, temp_disB, temp_disC, dis_to_pointA, dis_to_pointB, dis_to_pointC);
		if ((mag(point - intersection_point) <= MININT) &&
			(mag(point - intersection_point) >= - MININT))
			return - dis;

		if (dis == dis_to_pointA)
		{
			dis = dis * sign_A;
			return dis;
		}
		if (dis == dis_to_pointB)
		{
			dis = dis * sign_B;
			return dis;
		}
		if (dis == dis_to_pointC)
		{
			dis = dis * sign_C;
			return dis;
		}
		if (dis == temp_disA)
		{
			dis = dis * sign_edge_A;
			return dis;
		}
		if (dis == temp_disB)
		{
			dis = dis * sign_edge_B;
			return dis;
		}
		if(dis == temp_disC)
		{
			dis = dis * sign_edge_C;
			return dis;
		}
	}
}

float Potential::inside_or_outside(Vec3f point, Vec3f signormal)
{
	if (dot(point,signormal)>0)
	{
		return -1.0;
	}
	else
	{
		return 1.0;
	}
}
bool Potential::Is_point_on_segment(Vec3f point,Vec3f pointA,Vec3f pointB)
{
	if (dot(point-pointA,point-pointB)>0)
	{
		return false;
	}
	else
	{
		return true;
	}
}