//Potential.h

#pragma once

#ifndef _POTENTIAL_H_
#define _POTENTIAL_H_

#include "vec.h"
#include <queue>
#include "objloader.h"
using namespace std;

#define MAXgridX 40   
#define MAXgridY 40
#define	MAXgridZ 40
#define GRIDL 1
#define KNOWN 0    
#define CANDIDATE 1       
#define UNKNOWN 2
#define MININT 0.0001 
#define MAXINT 100000
#define DISVALUE 100;

struct Node{
	int x, y, z;
	float value;
};

struct cmp{
	bool operator() ( Node a, Node b ){
		return a.value > b.value;
	}
};

typedef struct{
	float speed;
	float cost;
	Vec3f velocity;
}faceDataStruct;

typedef struct{ 
	float discomfort; 
	float potential;
	faceDataStruct faceData[6]; 
	int FMMFlag;
}gridDataStruct;

class Potential 
{
public:

	//
	gridDataStruct m_gridData[MAXgridX][MAXgridY][MAXgridZ];
	//ܶȳȺܶȺϰγɵܶȳ
	float density[MAXgridX][MAXgridY][MAXgridZ]; 
	//ƽٶȳܶȳɣspeed
	Vec3f averagev[MAXgridX][MAXgridY][MAXgridZ];
	//ϰϰܶȳ
	float distance[MAXgridX][MAXgridY][MAXgridZ];

	float maxSpeed, minSpeed;
	int m_GoalXa, m_GoalXb, m_GoalYa, m_GoalYb, m_GoalZa, m_GoalZb;
	const int gridX, gridY, gridZ;

	float maxDensity, minDensity;
	float lamda;
	float w1, w2, w3;

public:

	Potential(int x,int y,int z);
	void SetGoal(int xa, int xb, int ya, int yb, int za, int zb);
	void SetSpeedRange(float _minSpeed, float _maxSpeed);
	void SetDensityRange(float _minDensity, float _maxDensity);
	void SetLamdaWeights(float _lamda, float _w1, float _w2, float _w3);
	void ComputeFaceData(void);
	void FMM(void);
	Vec3f GetVelocity(Vec3f position);
	void addDensity(Vec3f position,Vec3f velocity);
	Vec3f get_PotentialGradient(Vec3f position);
	Vec3f get_DensityGradient(Vec3f position);
	bool addObj(const char* filename);
	void ComputeDiscomfort(void);

private:

	void CalcPotential(int x, int y, int z);
	void solve(float &x, float a, float b, float c, float d);
	void process(int x, int y, int z, priority_queue<Node, vector<Node>, cmp>& Candidate);
	float ComputeDistance(objloader model, int trangle_index, Vec3f point);
	float inside_or_outside(Vec3f point, Vec3f signormal);
	bool Is_point_on_segment(Vec3f point,Vec3f pointA,Vec3f pointB);
	void Floodfill(int i, int j, int k);
	
};

#endif