// Two group Demo
// ----------------------------------------------------------------------------


// ----------------------------------------------------------------------------

#include "Potential.h"

#define STARTZ 1
#define Step 1.8
#define FD 15
#define DDIS 5 
#define INDEX 5

using namespace OpenSteer;
// ----------------------------------------------------------------------------
class Pedestrian0 : public SimpleVehicle
{
public:
	float m_mdlTime;
	float m_mdlStep;
	mdlobj * m_mdlobj;

	Pedestrian0 (int groupId,mdlobj *obj) 
	{
		GroupID = groupId;
		m_mdlobj = obj;
		reset ();

	}

	// reset state
	void reset (void)
	{
		m_index = 0;
		for(int i=0;i<INDEX;i++) m_speed[i]=Vec3.zero;
		SimpleVehicle::reset ();	
		const float rRadius = frandom2 (0, 1);
		const Vec3 randomOnRing = RandomUnitVectorOnXZPlane () * rRadius;
		Vec3 startPosition;
		switch(GroupID)
		{
		case 0:startPosition.set(-20+1+rand()%2,0,-10+StartID*Step);
			setForward(Vec3(1,0,0));
			break;
		case 1:startPosition.set(-20+39-rand()%2,0,-10+StartID*Step);
			setForward(Vec3(-1,0,0));
			break;
		}
		StartID++;
		if(StartID>10) StartID = 1;
		setRadius(0.7);
		setPosition (startPosition + randomOnRing);
		// speed along Forward direction.
		setSpeed (2.5);
		// steering force clip magnitude
		setMaxForce (0.3f);
		// velocity  clip magnitude
		setMaxSpeed (5.0f);
		// 15 seconds and 150 points along the trail
		clearTrailHistory ();
		setTrailParameters (15, 150);
		m_mdlTime = GetTickCount( )/1000.0f;
		m_mdlStep = 0;
	}

	// draw into the scene
	void draw ()
	{
		if(m_mdlobj)
		{
			drawMdlVehicle(*this,m_mdlTime,m_mdlStep,1,m_mdlobj,gGray50,0.07);
			switch(GroupID)
			{
			case 0:drawTrail (gGreen,gGreen);
				break;
			case 1:drawTrail (gRed,gRed);
				break;
			}
		}	
	}

	// per frame simulation update
	void update (const float currentTime, const float elapsedTime,TotalDataStruct &TotalData)
	{
		Vec3 p = position();
		int m = floor(p.x+20);
		int n = floor(p.z+10);
		//Ƹ岻
		if(m<0||m>=40||n<0||n>=20)
		{
			if(m<0) p.x = 0;
			if(m>=40) p.x = 39.9;
			if(n<0) p.z = 0;
			if(n>=20) p.z = 19.9;
			setPosition(p);
		}
		Vec3 front;
		switch(GroupID)
		{
		case 0:front.set(1,0,0);
			break;
		case 1:front.set(-1,0,0);
			break;
		}
		//õٶ
		Vec3 v = m_potential[GroupID].GetVelocity(position(),front,true,TotalData);
		m_speed[m_index] = v;
		m_index++;
		if(m_index>=INDEX) m_index = 0;
		Vec3 temp;
		for(int i = 0;i<INDEX;i++) temp += m_speed[i];
		temp = temp/INDEX;
		setSpeed(temp.length());
		if(temp.length()>=0.1) setForward(temp.normalize());
		applySteeringForce (temp,elapsedTime);

		//applySteeringForce (front,elapsedTime);
		// annotation
		//annotationVelocityAcceleration ();
		recordTrailVertex (currentTime, position());
	}
	void updateDensity(TotalDataStruct &TotalData)
	{
		Vec3 p = position();
		int m = floor((p.x+20)/GRIDL);
		int n = floor((p.z+10)/GRIDL);
		Vec3 v = velocity();
		m_potential[0].addDensity(p,v,TotalData);
		//ȺǰĲӦȣʹȺԱܿ
		switch(GroupID)
		{
		case 0:
			for(int i=0;i<DDIS&&m+i<40;i++)
				m_potential[1].m_gridData[m+i][n].discomfort = FD;
			break;
		case 1:
			for(int i=0;i<DDIS&&m-i>0;i++)
				m_potential[0].m_gridData[m-i][n].discomfort = FD;
			break;
		}

	}

	// for stepping the starting conditions for next vehicle
	int GroupID;
	Vec3 m_speed[INDEX];
	int m_index;
	static int StartID;
	static Potential m_potential[2];
};
int Pedestrian0::StartID = 1;
Potential Pedestrian0::m_potential[2] = {Potential(40,20,0.45,0.01,2,1,6,2), 
Potential(40,20,0.45,0.01,2,1,6,2)};

// ----------------------------------------------------------------------------
// PlugIn for OpenSteerDemo


const int lstCount = 10;


class TwoGroupPlugIn : public PlugIn
{
public:
	Model m_3ds;	//3ds̬ģ
	int gridX;
	int gridZ;
	mdlobj	m_mdlobj[2];  //ģ
	const char* name (void) {return "Two group";}

	float selectionOrderSortKey (void) {return 0.001f;}

	// be more "nice" to avoid a compiler warning
	virtual ~TwoGroupPlugIn() {}

	void open (void)
	{
		// create a given number of agents with stepped inital parameters,
		// store pointers to them in an array.
		//m_roadData = mdlobj::LoadBit("../map/map_street.bmp",&m_bit);
		frame = 0;
		for(int i=0;i<2;i++)
		{
			char *mdl_name = new char[25];
			sprintf(mdl_name,"../mdl/walk0%d.mdl",i);
			m_mdlobj[i].InitGL(mdl_name);
			m_mdlobj[i].z_rot3=0.0f;
			m_mdlobj[i].Models=1;
			delete []mdl_name;
		}
		for (int i = 0; i < lstCount; i++) 
			all.push_back (new Pedestrian0(0,&m_mdlobj[0]));
		for (int i = 0; i < lstCount; i++) 
			all.push_back (new Pedestrian0(1,&m_mdlobj[1]));

		//趨ʼĿĵأٶȷΧ¶ȷΧ
		gridX = Pedestrian0::m_potential[0].gridX;
		gridZ = Pedestrian0::m_potential[0].gridZ;
		TotalData.gridX = gridX;
		TotalData.gridZ = gridZ;
		for(int m=0;m<gridX;m++)
		{
			for(int n=0;n<gridZ;n++)
			{
				TotalData.height[m][n] = 0;
			}
		}
		Pedestrian0::m_potential[0].SetGoal(gridX-2,gridX-1,0,gridZ-1);
		Pedestrian0::m_potential[0].SetSpeedRange(2.5,5.0);
		Pedestrian0::m_potential[0].SetSlopeRange(-2,2);
		Pedestrian0::m_potential[1].SetGoal(0,1,0,gridZ-1);
		Pedestrian0::m_potential[1].SetSpeedRange(2.5,5.0);
		Pedestrian0::m_potential[1].SetSlopeRange(-2,2);
		// initial selected vehicle
		OpenSteerDemo::selectedVehicle = *all.begin();

		// initialize camera

		OpenSteerDemo::camera.setPosition (0, 0, 0);
		OpenSteerDemo::camera.fixedPosition.set (0, 40, 20);
		OpenSteerDemo::camera.mode = Camera::cmFixed;

		//OpenSteerDemo::camera.fixedPosition.set (20*GRIDL, 20*GRIDL, 15*GRIDL);
		//OpenSteerDemo::camera.mode = Camera::cmFixed;
		OpenSteerDemo::PlugInMsg.Format("");

		//for 3ds
		m_3ds.Init("../3ds/border","border.3ds");
		//m_3ds.Init("../3ds/butterfly","butterfly.3ds");
		GLfloat LightAmbient[]=	{ 0.5f, 0.5f, 0.5f, 1.0f };
		GLfloat LightDiffuse[]=	{ 1.0f, 1.0f, 1.0f, 1.0f };
		GLfloat LightPosition[]={ 1.0f, 1.0f, 1.0f, 0.0f };
		glLightfv(GL_LIGHT1, GL_AMBIENT, LightAmbient);	
		glLightfv(GL_LIGHT1, GL_DIFFUSE, LightDiffuse);	
		glLightfv(GL_LIGHT1, GL_POSITION,LightPosition);
		//end
	}

	void update (const float currentTime, const float elapsedTime)
	{
		bool flag1 = OpenSteer::OpenSteerDemo::clock.getPausedState();
		bool flag2 = OpenSteer::OpenSteerDemo::stepFlag;
		if(flag1&&!flag2) return;
		for(int m=0;m<gridX;m++)
			for(int n=0;n<gridZ;n++)
			{
				TotalData.density[m][n] = 0;
				TotalData.averagev[m][n] = Vec3(0,0,0);
			}
			resetDiscomford();
			//ۼƼTotalDataеܶȺٶ
			for (iterator i = all.begin(); i != all.end(); i++)
				(**i).updateDensity(TotalData);
			for(int m=0;m<gridX;m++)
				for(int n=0;n<gridZ;n++)
					if(TotalData.density[m][n]>0) 
						TotalData.averagev[m][n] = TotalData.averagev[m][n]/TotalData.density[m][n];
			//ܳ
			Pedestrian0::m_potential[0].ComputeFaceData(TotalData);
			Pedestrian0::m_potential[1].ComputeFaceData(TotalData);
			Pedestrian0::m_potential[0].FMM();
			Pedestrian0::m_potential[1].FMM();
			// update, draw and annotate each agent
			for (iterator i = all.begin(); i != all.end(); i++)
				(**i).update (currentTime, elapsedTime, TotalData);	
			int endFlag = 0;
			//һĿĸ嵽Ŀأreset
			for (iterator i = all.begin(); i != all.end(); i++)
			{
				Vec3 p = (**i).position();
				int m = floor(p.x+20);
				int n = floor(p.z+10);
				if(Pedestrian0::m_potential[(**i).GroupID].m_gridData[m][n].potential == 0 )
					endFlag ++;
			}
			if(endFlag>=16)
				for (iterator i = all.begin(); i != all.end(); i++)
					(**i).reset();

			//ײ
			for (iterator i = all.begin(); i != all.end(); i++)
				for (iterator j = i+1; j != all.end(); j++)
				{		
					float dist = Vec3::distance((**i).position(),(**j).position());
					if(dist< ((**i).radius()+(**j).radius()))
					{
						Vec3 fij = (**j).position()-(**i).position();
						Vec3 adjust = fij.normalize()*((**i).radius()+(**j).radius());
						adjust = (adjust - fij)/2.0;
						(**i).setPosition((**i).position()-adjust);
						(**j).setPosition((**j).position()+adjust);						
					}
				}
	}
	void redraw (const float currentTime, const float elapsedTime)
	{
		// selected vehicle (user can mouse click to select another)
		AbstractVehicle& selected = *OpenSteerDemo::selectedVehicle;

		// vehicle nearest mouse (to be highlighted)
		AbstractVehicle& nearMouse = *OpenSteerDemo::vehicleNearestToMouse ();

		// update camera
		OpenSteerDemo::updateCamera (currentTime, elapsedTime, selected);


		// draw "ground plane"
		const Vec3 gridCenter (0,0,0);
		// draw checkerboard grid with 50 squares along each side
		drawXZCheckerboardGrid2 (2.0,gridX*GRIDL, gridZ*GRIDL, gridCenter, grayColor (0.7f), grayColor (0.9f));

		/*
		glPushAttrib (GL_ENABLE_BIT);
		glDisable (GL_CULL_FACE);
		{
		for(int m=0;m<gridX;m++)
		for(int n=0;n<gridZ;n++)
		{
		//float color = Pedestrian0::m_potential[0].m_gridData[m][n].faceData[2].speed/10;
		//float color = Pedestrian0::m_potential[1].m_gridData[m][n].potential/50.0;
		//float color = TotalData.averagev[m][n].length()/2.0; 
		//float color = TotalData.density[m][n]/1.0;
		if(color>1) color = 1;
		glBegin(GL_QUADS);
		glColor3f(color,color,color);
		glVertex3f(m*GRIDL,0,n*GRIDL);
		glVertex3f((m+1)*GRIDL,0,n*GRIDL);
		glVertex3f((m+1)*GRIDL,0,(n+1)*GRIDL);
		glVertex3f(m*GRIDL,0,(n+1)*GRIDL);
		glEnd();
		}
		}
		glPopAttrib ();
		*/	

		// update, draw and annotate each agent
		for (iterator i = all.begin(); i != all.end(); i++)
		{
			// draw this agent
			Pedestrian0& agent = **i;
			agent.draw ();
		}
		// highlight vehicle nearest mouse
		OpenSteerDemo::highlightVehicleUtility (nearMouse);

		//for 3ds
		glPushAttrib (GL_ENABLE_BIT|GL_LIGHTING_BIT|GL_TEXTURE_BIT);
		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT1);
		glDisable (GL_CULL_FACE);
		m_3ds.show3ds(0,0,0,1,Vec3(1,0,0),Vec3(0,1,0),Vec3(0,0,1));
		glPopAttrib();
		//end
	}

	void close (void)
	{
		for (iterator i = all.begin(); i!=all.end(); i++) delete (*i);
		all.clear ();
		for(int i=0;i<2;i++)
		{
			m_mdlobj[i].FreeGL();
		}
	}

	void reset (void)
	{
		// reset each agent
		for (iterator i = all.begin(); i!=all.end(); i++) (**i).reset();
	}
	void resetDiscomford(void)
	{
		for(int m=0;m<gridX;m++)
		{
			for(int n=0;n<gridZ;n++)
			{
				Pedestrian0::m_potential[0].m_gridData[m][n].discomfort = 0;
				Pedestrian0::m_potential[1].m_gridData[m][n].discomfort = 0;
			}
		}
	}
	const AVGroup& allVehicles (void) {return (const AVGroup&) all;}

	std::vector<Pedestrian0*> all;
	typedef std::vector<Pedestrian0*>::const_iterator iterator;
	TotalDataStruct TotalData;
	int frame;
};


TwoGroupPlugIn gTwoGroupPlugIn;
