#ifndef RVO_AGENT_H
#define RVO_AGENT_H


#include "stdafx.h"
#include "Simulator.h"

namespace RVO
{
/*!
 *  @brief		Defines an agent in the simulation.
 */
class Agent
{
private:
	/*!
	 *  @brief		Constructs an agent instance.
	 *  @param		sim		The simulator instance.
	 */
	explicit Agent(Simulator* sim);

	/*!
	 *  @brief		Computes the neighbors of this agent.
	 */
	void computeNeighbors();

	/*!
	 *  @brief		Computes the new velocity of this agent.
	 */
	void computeNewVelocity();

	/*!
	 *  @brief		Inserts an agent neighbor into the set of neighbors of this agent.
	 *  @param		agent		A pointer to the agent to be inserted.
	 *  @param		rangeSq		The squared range around this agent.
	 */
	void insertAgentNeighbor(const Agent* agent, float& rangeSq);

	/*!
	 *  @brief		Updates the three-dimensional position and three-dimensional velocity of this agent.
	 */
	void update();

	H3DNode Node;		//horde3dʶ
	float AnimTime;  //󶨵ĶϢ

	std::vector<std::pair<float, const Agent*> > agentNeighbors_;
	size_t id_;
	size_t maxNeighbors_;
	float maxSpeed_;
	float neighborDist_;
	Vec3f newVelocity_;
	std::vector<Plane> orcaPlanes_;
	Vec3f position_;
	Vec3f prefVelocity_;
	float radius_;
	Simulator* sim_;
	float timeHorizon_;
	Vec3f velocity_;
	int statement;

	friend class KdTree;
	friend class Simulator;
};
}

#endif
